#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import combinations, product
from typing import TypedDict
import logging

from nnf import Var, And

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class ChoiQUBO(Mapping):
    """
    QUBO formulation for SAT problem by Choi (1004.2226).
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]
        self.nr_vars = None
        self.reverse_dict = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "nnf", "version": "0.4.1"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary with parameter options
        .. code-block:: python

            return {
                    "hard_reward": {
                        "values": [0.1, 0.5, 0.9, 0.99],
                        "description": "What Bh/A ratio do you want? (How strongly to enforce hard constraints)"
                    },
                    "soft_reward": {
                        "values": [0.1, 1, 2],
                        "description": "What Bh/Bs ratio do you want? This value is multiplied with the "
                        "number of tests."
                    }
                }
        """
        return {
            "hard_reward": {
                "values": [0.1, 0.5, 0.9, 0.99],
                "description": (
                    "What Bh/A ratio do you want?"
                    "(How strongly to enforce hard constraints)"
                )
            },
            "soft_reward": {
                "values": [0.1, 1, 2],
                "description": (
                    "What Bh/Bs ratio do you want?"
                    "This value is multiplied with the number of tests."
                )
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             hard_reward: float
             soft_reward: float

        """
        hard_reward: float
        soft_reward: float

    def map(self, problem: tuple[And, list], config: Config) -> tuple[dict, float]:
        """
        Converts a MaxSAT instance with hard and soft constraints into a graph problem --
        solving MaxSAT then corresponds to solving an instance of the Maximal Independent Set problem.
        See Andrew Lucas (2014), or the original publication by Choi (1004.2226).

        :param problem: A tuple containing hard and soft constraints
        :param config: Config with the parameters specified in Config class
        :return: Dictionary containing the QUBO representation and the time taken
        """
        start = start_time_measurement()

        hard_constraints, soft_constraints = problem
        a = 1
        bh = config['hard_reward'] * a
        # divide Bh by the number of test clauses, such that fulfilling a test result is less favourable than
        # satisfying a constraint, which aim to prioritize.
        bs = bh * config['soft_reward'] / len(soft_constraints)

        # Count the number of different variables that appear in the vehicle options problem:
        self.nr_vars = len(hard_constraints.vars().union(And(soft_constraints).vars()))
        # Edges variable holds all edges in the resulting graph
        edges = {}
        # lit_occur is a dictionary which will store the information in which clause a certain literal will occur.
        lit_occur = {}

        def _add_clause(clause, curr_edges, curr_lit_occ, pos):
            literals = [f"{el}-{pos}" for el in clause.children]
            # Connect the literals within one clause
            for cmb in combinations(literals, 2):
                # Add a weight for each edge within clause
                curr_edges[cmb] = a
            # Add the occurrences of the variables to the occurrences dictionary
            for var in clause.children:
                if var.name not in curr_lit_occ.keys():
                    curr_lit_occ[var.name] = {True: [], False: []}
                # Add occurrences and mark that they correspond to hard constraints
                curr_lit_occ[var.name][var.true].append(pos)
            return curr_edges, curr_lit_occ

        # Convert the hard constraints into the graph
        for idx, hard_constraint in enumerate(hard_constraints):
            edges, lit_occur = _add_clause(hard_constraint, edges, lit_occur, idx)

        # Save the current total clause count:
        constraints_max_ind = len(hard_constraints)
        # Repeat the procedure for the soft constraints:
        for idx, soft_constraint in enumerate(soft_constraints):
            edges, lit_occur = _add_clause(soft_constraint, edges, lit_occur, idx + constraints_max_ind)

        # Connect conflicting clauses using the lit_occur dict:
        for literal, positions_dict in lit_occur.items():
            # for every literal lit, we check its occurrences and connect the non-negated and negated occurrences.
            for pos_true, pos_false in product(positions_dict[True], positions_dict[False]):
                if pos_true != pos_false:
                    # Employ the notation from nnf, where the tilde symbol ~ corresponds to negation.
                    lit_true, lit_false = f"{literal}-{pos_true}", f"~{literal}-{pos_false}"
                    # Add a penalty for each such edge:
                    edges[(lit_true, lit_false)] = a

        # Collect all different nodes that we have in our graph, omitting repetitions:
        node_set = set([])
        for nodes in edges.keys():
            node_set = node_set.union(set(nodes))

        node_list = sorted(node_set)
        # Fix a mapping (node -> binary variable)
        relabel_dict = {v: i for i, v in enumerate(node_list)}
        # Save the reverse mapping, which is later used to decode the solution.
        self.reverse_dict = dict(enumerate(node_list))

        def _remap_pair(pair):
            """Small helper function that maps the nodes of an edge to binary variables"""
            return relabel_dict[pair[0]], relabel_dict[pair[1]]

        # Save the QUBO corresponding to the graph.
        q = {_remap_pair(key): val for key, val in edges.items()}

        for v in node_list:
            # Add different energy rewards depending on whether it is a hard or a soft constraint
            if int(v.split('-')[-1]) < constraints_max_ind:
                # if hard cons, add -Bh as the reward
                q[_remap_pair((v, v))] = -bh
            else:
                # for soft constraints, add -Bs
                q[_remap_pair((v, v))] = -bs

        logging.info(f"Converted to Choi QUBO with {len(node_list)} binary variables. Bh={config['hard_reward']},"
                     f" Bs={bs}.")
        return {'Q': q}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the SAT class for validation/evaluation.

        :param solution: Dictionary containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()
        # We define the literals list, so that we can check the self-consistency of the solution. That is, we save all
        # assignments proposed by the annealer, and see if there is no contradiction. (In principle a solver
        # could mandate L3 = True and L3 = False, resulting in a contradiction.)
        literals = []
        # Assignments saves the actual solution
        assignments = []

        for node, tf in solution.items():
            # Check if node is included in the set (i.e. if tf is True (1))
            if tf:
                # Convert back to the language of literals
                lit_str = self.reverse_dict[node]
                # Check if the literal is negated:
                if lit_str.startswith('~'):
                    # Remove the negation symbol
                    lit_str = lit_str.replace('~', '')
                    # Save a negated literal object, will be used for self-consistency check
                    lit = Var(lit_str).negate()
                    # Add the negated literal to the assignments, removing the (irrelevant) position part
                    assignments.append(Var(lit_str.split('-')[0]).negate())
                else:
                    # If literal is true, no ~ symbol needs to be removed:
                    lit = Var(lit_str)
                    assignments.append(Var(lit_str.split('-')[0]))
                literals.append(lit)

        # Check for self-consistency of solution; Check that the assignments of all literals are consistent:
        if not And(set(literals)).satisfiable():
            logging.error('Generated solution is not self-consistent!')
            raise ValueError("Inconsistent solution for the ChoiQubo returned.")

        # If the solution is consistent, find and add potentially missing variables:
        assignments = sorted(set(assignments))
        # Find missing vars, or more precisely, their labels:
        missing_vars = set(range(self.nr_vars)) - {int(str(a).replace('L', '').replace('~', '')) for a in assignments}

        # Add the variables that found were missing:
        for nr in missing_vars:
            assignments.append(Var(f'L{nr}'))

        return {list(v.vars())[0]: v.true for v in sorted(assignments)}, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from braket.aws import AwsDevice

from quark.modules.devices.braket.Braket import Braket


class Rigetti(Braket):
    """
    Class for using the Rigetti devices on Amazon Braket.
    """

    def __init__(self, device_name: str, arn: str = 'arn:aws:braket:us-west-1::device/qpu/rigetti/Ankaa-2'):
        """
        Constructor method.

        :param device_name: Name of the device
        :param arn: Amazon resource name for the Rigetti device
        """
        super().__init__(region="us-west-1", device_name=device_name, arn=arn)
        self.submodule_options = []

        if 'SKIP_INIT' in os.environ:
            # TODO: This is currently needed so create_module_db in the Installer does not execute the rest
            # of this section, which would be unnecessary. However, this should be done better in the future!
            return

        self.init_s3_storage("rigetti")
        self.device = AwsDevice(arn, aws_session=self.aws_session)

    def get_parameter_options(self) -> dict:
        """
        Returns an empty dictionary as this solver has no configurable settings.

        :return: Empty dict
        """
        return {}

    def get_default_submodule(self, option: str) -> None:
        """
        Raises ValueError as this module has no submodules.

        :param option: Option name
        :raises ValueError: If called, since this module has no submodules
        """
        raise ValueError("This module has no submodules.")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import logging


def is_running_mpiexec():
    """
    Determines if the script is running under mpiexec.

    :return: True if running under mpiexec, False otherwise
    :rtype: bool
    """
    # This is not 100% robust but should cover MPICH & Open MPI
    for key in os.environ:
        if key.startswith("PMI_") or key.startswith("OMPI_COMM_WORLD_"):
            return True
    return False


def is_running_mpi():
    """
    Determines if the MPI environment is available and import mpi4py if so.

    :return: MPI object if available, None otherwise
    :rtype: MPI or None
    """
    if is_running_mpiexec():
        try:
            from mpi4py import MPI  # pylint: disable=C0415
        except ImportError as e:
            raise RuntimeError(
                'it seems you are running mpiexec/mpirun but mpi4py cannot be '
                'imported, maybe you forgot to install it?'
            ) from e
    else:
        MPI = None
    return MPI


class MPIStreamHandler(logging.StreamHandler):
    """
    A logging handler that only emits records from the root process in an MPI environment.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        MPI = is_running_mpi()
        self.rank = MPI.COMM_WORLD.Get_rank() if MPI else 0

    def emit(self, record) -> None:
        """
        Emits a log record if running on the root process.

        :param record: Log record
        :type record: Logging.LOgRecord
        """
        if self.rank == 0:
            super().emit(record)


class MPIFileHandler(logging.FileHandler):
    """
    A logging handler that only emits records to a file from the root process in an MPI environment.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        MPI = is_running_mpi()
        self.rank = MPI.COMM_WORLD.Get_rank() if MPI else 0

    def emit(self, record) -> None:
        """
        Emits a log record if running on the root process.

        :param record: Log record
        :type record: Logging.LOgRecord
        """
        if self.rank == 0:
            super().emit(record)


def get_comm() -> any:
    """
    Retrieves the MPI communicator if running in an MPI environment, otherwise provides a mock comm class.

    return: MPI communicator or a mock class with limited methods
    """
    mpi = is_running_mpi()
    if mpi:
        Comm = mpi.COMM_WORLD
    else:
        class Comm():
            @staticmethod
            def Get_rank():
                return 0

            @staticmethod
            def Bcast(loss, root):
                pass

            @staticmethod
            def Barrier():
                pass

    return Comm



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
import logging

import matplotlib
from matplotlib import pyplot as plt
import seaborn as sns
import pandas as pd

matplotlib.use('Agg')
matplotlib.rcParams['savefig.dpi'] = 300
sns.set(style="darkgrid")


class Plotter:
    """
    Plotter class which generates some general plots.
    """

    @staticmethod
    def visualize_results(results: list[dict], store_dir: str) -> None:
        """
        Function to plot the execution times of the benchmark.

        :param results: Dict containing the results
        :param store_dir: Directory where the plots are stored
        """
        if results is None or len(results) == 0:
            logging.info("Nothing to plot since results are empty.")
            return

        processed_results_with_application_score = []
        processed_results_rest = []
        required_application_score_keys = [
            "application_score_value", "application_score_unit", "application_score_type"
        ]
        application_name = None
        application_axis = None
        static_keys, changing_keys = Plotter._get_config_keys(results)
        for result in results:
            application_name = result["module"]["module_name"]
            if len(changing_keys) == 1:
                # If only 1 config item changes, we use its value for application_config
                application_axis = changing_keys[0]
                application_config = result['module']['module_config'][changing_keys[0]]
            else:
                # If multiple config items change, we stringify them
                application_axis = f"{application_name} Config"
                application_config = ', '.join(
                    [f"{key}: {value}" for (key, value) in sorted(result["module"]["module_config"].items(),
                                                                  key=lambda key_value_pair:
                                                                  key_value_pair[0]) if key not in static_keys]
                )
            if len(static_keys) > 0:
                # Include the static items in the axis name
                application_axis += "(" + ', '.join(
                    [f"{key}: {result['module']['module_config'][key]}" for key in static_keys]
                ) + ")"

            processed_item = Plotter._extract_columns({
                "benchmark_backlog_item_number": result["benchmark_backlog_item_number"],
                "total_time": result["total_time"],
                "application_config": application_config
            }, result["module"])

            if all(k in result["module"] for k in required_application_score_keys):
                # Check if all required keys are present to create application score plots
                for k in required_application_score_keys:
                    processed_item[k] = result["module"][k]
                processed_results_with_application_score.append(processed_item)
            else:
                processed_results_rest.append(processed_item)

        if len(processed_results_with_application_score) > 0:
            logging.info("Found results with an application score, generating according plots.")
            Plotter.plot_application_score(
                application_name, application_axis, processed_results_with_application_score, store_dir
            )

        Plotter.plot_times(
            application_name, application_axis,
            [*processed_results_with_application_score, *processed_results_rest],
            store_dir, required_application_score_keys
        )

        logging.info("Finished creating plots.")

    @staticmethod
    def plot_times(application_name: str, application_axis: str, results: list[dict], store_dir: str,
                   required_application_score_keys: list) -> None:
        """
        Function to plot execution times of the different modules in a benchmark.

        :param application_name: Name of the application
        :param application_axis: Name of the application axis
        :param results: Dict containing the results
        :param store_dir: Directory where the plots are stored
        :param required_application_score_keys: List of keys which have to be present to calculate an application score
        """

        df = pd.DataFrame.from_dict(results)
        df = df.fillna(0.0).infer_objects(copy=False)
        df_melt = df.drop(df.filter(["application_config", "config_combo", "total_time",
                                     *required_application_score_keys]), axis=1)
        df_melt = pd.melt(frame=df_melt, id_vars='benchmark_backlog_item_number', var_name='module_config',
                          value_name='time')

        # This plot shows the execution time of each module
        ax = sns.barplot(x="benchmark_backlog_item_number", y="time", data=df_melt, hue="module_config")
        plt.title(application_name)
        # Put the legend out of the figure
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., title="Modules used")
        ax.set(xlabel="benchmark run ID", ylabel='execution time of module (ms)')
        plt.savefig(f"{store_dir}/time_by_module.pdf", dpi=300, bbox_inches='tight')
        logging.info(f"Saved {f'{store_dir}/time_by_module.pdf'}.")
        plt.clf()

        # This plot shows the total time of a benchmark run
        ax = sns.barplot(x="application_config", y="total_time", data=df, hue="config_combo")
        ax.set(xlabel=application_axis, ylabel="total execution time (ms)")
        # Put the legend out of the figure
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., title="Modules used")
        plt.title(application_name)
        plt.sca(ax)
        # If column values are very long and of type string rotate the ticks
        if (pd.api.types.is_string_dtype(df.application_config.dtype) or pd.api.types.is_object_dtype(
                df.application_config.dtype)) and df.application_config.str.len().max() > 10:
            plt.xticks(rotation=90)
        plt.savefig(f"{store_dir}/total_time.pdf", dpi=300, bbox_inches='tight')
        logging.info(f"Saved {f'{store_dir}/total_time.pdf'}.")
        plt.clf()

    @staticmethod
    def plot_application_score(application_name: str, application_axis: str, results: list[dict],
                               store_dir: str) -> None:
        """
        Function to create plots showing the application score.

        :param application_name: Name of the application
        :param application_axis: Name of the application axis
        :param results: Dict containing the results
        :param store_dir: Directory where the plots are stored
        """
        df = pd.DataFrame.from_dict(results)
        application_score_units = df["application_score_unit"].unique()
        count_invalid_rows = pd.isna(df['application_score_value']).sum()

        if count_invalid_rows == len(df):
            logging.info("All results have an invalid application score, skipping plotting.")
            return
        else:
            logging.info(f"{count_invalid_rows} out of {len(df)} benchmark runs have an invalid application score.")

        if len(application_score_units) != 1:
            logging.warning(
                f"Found more or less than exactly 1 application_score_unit in {application_score_units}."
                f" This might lead to incorrect plots!"
            )

        ax = sns.barplot(x="application_config", y="application_score_value", data=df, hue="config_combo")
        ax.set(xlabel=application_axis, ylabel=application_score_units[0])
        # Put the legend out of the figure
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., title="Modules used")
        ax.text(
            1.03, 0.5,
            f"{len(df) - count_invalid_rows}/{len(df)} runs have a valid \napplication score",
            transform=ax.transAxes, fontsize=12, verticalalignment='top',
            bbox={"boxstyle": "round", "alpha": 0.15}
        )
        plt.title(application_name)

        plt.sca(ax)
        # If column values are very long and of type string, rotate the ticks
        if (pd.api.types.is_string_dtype(df.application_config.dtype) or pd.api.types.is_object_dtype(
                df.application_config.dtype)) and df.application_config.str.len().max() > 10:
            plt.xticks(rotation=90)

        plt.savefig(f"{store_dir}/application_score.pdf", dpi=300, bbox_inches='tight')
        logging.info(f"Saved {f'{store_dir}/application_score.pdf'}.")
        plt.clf()

    @staticmethod
    def _get_config_keys(results: list[dict]) -> tuple[list, list]:
        """
        Function that extracts config keys.

        :param results: Results of a benchmark run
        :return: Tuple with list of static keys and list of changing keys
        """
        static_keys = []
        changing_keys = []
        helper_dict = defaultdict(list)
        # Try to find out which key in the dict change
        for result in results:  # you can list as many input dicts as you want here
            d = result["module"]["module_config"]
            for key, value in d.items():
                helper_dict[key].append(value)
                helper_dict[key] = list(set(helper_dict[key]))

        for key, value in helper_dict.items():
            if len(value) == 1:
                static_keys.append(key)
            else:
                changing_keys.append(key)

        return static_keys, changing_keys

    @staticmethod
    def _extract_columns(config: dict, rest_result: dict) -> dict:
        """
        Function to extract and summarize certain data fields like the time spent in every module
        from the nested module chain.

        :param config: Dictionary containing multiple data fields like the config of a module
        :param rest_result: Rest of the module chain
        :return: Extracted data
        """
        if rest_result:
            module_name = rest_result["module_name"]
            for key, value in sorted(rest_result["module_config"].items(),
                                     key=lambda key_value_pair: key_value_pair[0]):
                module_name += f", {key}: {value}"

            config_combo = config.pop("config_combo") + "\n" + module_name if "config_combo" in config else ""
            return Plotter._extract_columns(
                {
                    **config,
                    "config_combo": config_combo,
                    module_name: rest_result["total_time"]
                    if module_name not in config else config[module_name] + rest_result["total_time"]
                },
                rest_result["submodule"]
            )

        return config



#  Copyright 2022 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module containing the ACL

This module initializes the ACL application, which is responsible for formulating
and solving the ACL problem using various mappings and solvers.
"""



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TypedDict

from qubovert.problems import SetCover
from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class QubovertQUBO(Mapping):
    """
    Qubovert formulation of the vehicle-options problem.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "qubovert", "version": "1.2.5"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary containing configurable settings
        .. code-block:: python

            return {
                "penalty_weight": {
                    "values": [2, 5, 10, 25, 50, 100],
                    "custom_input": True,
                    "custom_range": True,
                    "postproc": float,
                    "description": "Please choose the weight of the penalties in the QUBO representation of
                    the problem"
                }
            }
        """
        return {
            "penalty_weight": {
                "values": [2, 5, 10, 25, 50, 100],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": float,
                "description": "Please choose the weight of the penalties in the QUBO representation of the problem"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             penalty_weight: float

        """
        penalty_weight: float

    def map(self, problem: tuple, config: Config) -> tuple[dict, float]:
        """
        Maps the SCP to a QUBO matrix.

        :param problem: Tuple containing the set of all elements of an instance and a list of subsets,
                        each covering some of these elements
        :param config: Config with the parameters specified in Config class
        :return: Dict with QUBO matrix, time it took to map it
        """
        start = start_time_measurement()
        penalty_weight = config['penalty_weight']

        u, v = problem

        self.SCP_problem = SetCover(u, v)  # pylint: disable=W0201
        self.SCP_qubo = self.SCP_problem.to_qubo(penalty_weight)  # pylint: disable=W0201

        logging.info(f"Converted to QUBO with {self.SCP_qubo.num_binary_variables} Variables.")

        # Convert it to the right format to be accepted by Braket / Dwave
        q_dict = {}

        for key, val in self.SCP_qubo.items():
            # Interaction (quadratic) terms
            if len(key) == 2:
                if (key[0], key[1]) not in q_dict:
                    q_dict[(key[0], key[1])] = float(val)
                else:
                    q_dict[(key[0], key[1])] += float(val)
            # Local (linear) fields
            elif len(key) == 1:
                if (key[0], key[0]) not in q_dict:
                    q_dict[(key[0], key[0])] = float(val)
                else:
                    q_dict[(key[0], key[0])] += float(val)

        return {"Q": q_dict}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[set, float]:
        """
        Maps the solution of the QUBO to a set of subsets included in the solution.

        :param solution: QUBO matrix in dict form
        :return: Tuple with set of subsets that are part of the solution and the time it took to map it
        """
        start = start_time_measurement()
        sol = self.SCP_problem.convert_solution(solution)

        return sol, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from typing import TypedDict

from quark.modules.Core import Core
from quark.modules.devices.Device import Device


class QrispSimulator(Device, ABC):
    """
    Abstract class to use the Qrisp Simulator.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__(device_name="qrisp_simulator")
        self.submodule_options = []
        self.backend = None

    def get_backend(self) -> any:
        """
        Returns backend.

        :return: Instance of the backend class
        """
        if self.backend is None:
            raise AttributeError("The 'backend' attribute has not been set.")
        return self.backend

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "qrisp", "version": "0.5.2"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application, not Applicable for Qrisp Simulator for now.

        Example: "doppler": {
                "values": [False, True],
                "description": "Simulate doppler noise? Has a large impact on performance!"
            }
        """
        # TODO once optional noisy simulation is done in qrisp
        return {}

    class Config(TypedDict):
        """
        Attributes of a valid config.
        """
        doppler: bool

    def get_backend_config(self):
        """
        Returns backend configurations.

        :return: Backend config for the emulator
        """
        if self.backend is None:
            raise AttributeError("The 'backend' attribute has not been set.")
        if not hasattr(self.backend, 'config'):
            raise AttributeError("The 'backend' object has no attribute 'config'.")
        return self.backend.config

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the given option.

        :param option: The submodule option to select
        :return: Instance of the selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        raise ValueError("This module has no submodules.")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union, TypedDict

from quark.modules.applications.qml.generative_modeling.circuits.CircuitGenerative import CircuitGenerative
from quark.modules.applications.qml.generative_modeling.mappings.LibraryQiskit import LibraryQiskit
from quark.modules.applications.qml.generative_modeling.mappings.LibraryPennylane import LibraryPennylane
from quark.modules.applications.qml.generative_modeling.mappings.PresetQiskitNoisyBackend import PresetQiskitNoisyBackend
from quark.modules.applications.qml.generative_modeling.mappings.CustomQiskitNoisyBackend import CustomQiskitNoisyBackend


class CircuitStandard(CircuitGenerative):
    """
    This class generates a library-agnostic gate sequence, i.e. a list containing information
    about the gates and the wires they act on.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("DiscreteStandard")
        self.submodule_options = [
            "LibraryQiskit",
            "LibraryPennylane",
            "CustomQiskitNoisyBackend",
            "PresetQiskitNoisyBackend"
        ]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return []

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this standard circuit.

        :return: Dictionary of parameter options.
        .. code-block:: python

            return {
                    "depth": {
                        "values": [1, 2, 3],
                        "description": "What depth do you want?"
                    }
                }

        """
        return {

            "depth": {
                "values": [1, 2, 3],
                "description": "What depth do you want?"
            }
        }

    def get_default_submodule(self, option: str) \
            -> Union[LibraryQiskit, LibraryPennylane, PresetQiskitNoisyBackend, CustomQiskitNoisyBackend]:
        """
        Returns the default submodule based on the given option.

        :param option: The submodule option to select
        :return: Instance of the selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "LibraryQiskit":
            return LibraryQiskit()
        if option == "LibraryPennylane":
            return LibraryPennylane()
        elif option == "PresetQiskitNoisyBackend":
            return PresetQiskitNoisyBackend()
        elif option == "CustomQiskitNoisyBackend":
            return CustomQiskitNoisyBackend()
        else:
            raise NotImplementedError(f"Option {option} not implemented")

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             depth: int

        """
        depth: int

    def generate_gate_sequence(self, input_data: dict, config: Config) -> dict:
        """
        Returns gate sequence of standard architecture.

        :param input_data: Collection of information from the previous modules
        :param config: Config specifying the number of qubits of the circuit
        :return: Dictionary including the gate sequence of the Standard Circuit
        """
        n_registers = input_data["n_registers"]
        n_qubits = input_data["n_qubits"]
        depth = config["depth"]

        gate_sequence = []

        for _ in range(depth):
            for k in range(n_qubits):
                gate_sequence.append(["RY", [k]])

            for k in range(n_qubits - 1):
                gate_sequence.append(["RYY", [k, k + 1]])

            for k in range(n_qubits - 1):
                gate_sequence.append(["CRY", [k, k + 1]])

            gate_sequence.append(["Barrier", None])

        for k in range(n_qubits):
            gate_sequence.append(["Measure", [k, k]])

        output_dict = {
            "gate_sequence": gate_sequence,
            "circuit_name": "Standard",
            "n_qubits": n_qubits,
            "n_registers": n_registers,
            "depth": depth,
            "histogram_train": input_data["histogram_train"],
            "store_dir_iter": input_data["store_dir_iter"],
            "train_size": input_data["train_size"],
            "dataset_name": input_data["dataset_name"],
            "binary_train": input_data["binary_train"]
        }

        return output_dict



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# copyright Pulp:
# Copyright (c) 2002-2005, Jean-Sebastien Roy
# Modifications Copyright (c) 2007- Stuart Anthony Mitchell
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

import os
import logging
from typing import TypedDict

import pandas as pd
import numpy as np
import pulp

from quark.modules.applications.Application import Core
from quark.modules.applications.optimization.Optimization import Optimization
from quark.utils import start_time_measurement, end_time_measurement


class ACL(Optimization):
    """
    The distribution of passenger vehicles is a complex task and a high cost factor for automotive original
    equipment manufacturers (OEMs). Vehicles travel long distance on different carriers, such as ships,
    trains, and trucks, from the production plant to the customer.

    To save costs, OEMs and logistics service providers aim to maximize their loading capacities.
    Modern auto carriers are flexible, allowing individual platforms to be rotated, extended, or combined
    to accommodate vehicles of different shapes and weights in a space-efficient manner.

    In practice, finding feasible combinations is often based on heuristics or personal experience.
    We formulate the problem as a mixed integer quadratically constrained assignment problem.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("ACL")
        self.submodule_options = ["MIPsolverACL", "QUBO"]
        self.application = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns the list of module requirements.

        :return: List of dictionaries containing module requirements
        """
        return [
            {"name": "pulp", "version": "2.9.0"},
            {"name": "pandas", "version": "2.2.3"},
            {"name": "numpy", "version": "1.26.4"},
            {"name": "openpyxl", "version": "3.1.5"},
        ]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "MIPsolverACL":
            from quark.modules.solvers.MIPsolverACL import MIPaclp  # pylint: disable=C0415
            return MIPaclp()
        elif option == "QUBO":
            from quark.modules.applications.optimization.ACL.mappings.QUBO import Qubo  # pylint: disable=C0415
            return Qubo()
        else:
            raise NotImplementedError(f"Submodule Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns parameter options for selecting different models.

        :return: Dictionary containing model selection options
        """
        return {
            "model_select": {
                "values": ["Full", "Small", "Tiny"],
                "description": "Do you want the full model or a simplified (QUBO friendly) one?"
            }
        }

    class Config(TypedDict):
        model_select: str

    @staticmethod
    def intersectset(p1: list, p2: list) -> list:
        """
        Computes the intersection of two lists.

        :param p1: First list
        :param p2: Second list
        :return: List containing elements common to both p1 and p2
        """
        return np.intersect1d(p1, p2).tolist()

    @staticmethod
    def diffset(p1: list, p2: list) -> list:
        """
        Computes the difference between two lists.

        :param p1: First list
        :param p2: Second list
        :return: List containing elements in p1 that are not in p2
        """
        return np.setdiff1d(p1, p2).tolist()

    def generate_problem(self, config: Config) -> dict:  # pylint: disable=R0915
        """
        This function includes three models: Full, small and tiny. Full refers to the original model with all of its
        constraints. Small refers to the simplified model suitable for solving it with QC methods.
        The simplified model does not consider the possibility that vehicles can be angled or that they can be
        oriented forwards or backwards in relation to the auto carrier. For the tiny version we do not consider split
        platforms and consider only weight constraints.

        :param config: Config containing the selected scenario
        :return: Dictionary with scenario-dependent model formulated as linear problem
        """
        # Enter vehicles to load (BMW model codes)
        vehicles = ["G20", "G20", "G20", "G20", "G07", "G20"]
        # Import vehicle data
        df = pd.read_excel(os.path.join(os.path.dirname(__file__), "Vehicle_data_QUARK.xlsx"))
        model_select = config['model_select']

        # All the parameters are in decimeters and decitons (4m == 400 cm == 40 dm , 2 tons -> 20 dt
        # Below are the model specific parameters, constraints and objectives for the tiny, small and the full model

        if model_select == "Tiny":
            self._generate_tiny_model(df, vehicles)
        elif model_select == "Small":
            self._generate_small_model(df, vehicles)
        else:
            self._generate_full_model(df, vehicles)

        problem_instance = self.application.to_dict()
        self.application = problem_instance
        return self.application

    def _generate_tiny_model(self, df: any, vehicles: list) -> None:
        """
        Generate the problem model for the Tiny configurations.

        :param df: Datafile
        :param vehicles: List of vehicle types
        """
        # Weight parameters
        # Max. total weight on truck / trailer
        wt = [100]
        # Max. weight on the four levels
        wl = [50, 60]
        # Max. weights on platforms p, if not angled
        wp = [23, 23, 23, 26, 17]

        # Create empty lists for different vehicle parameters. This is required for proper indexing in the model.
        weight_list = [0] * (len(vehicles))

        for i, vehicle in enumerate(vehicles):
            df_new = df.loc[df['Type'] == vehicle]
            weight_list[i] = int(df_new["Weight"].iloc[0])

        # Set of available cars
        vecs = set(range(len(vehicles)))
        # Set of available platforms
        platforms_array = np.array([0, 1, 2, 3, 4])
        plats = set(range(len(platforms_array)))

        # Set of platforms that have a limitation on allowed weight
        platforms_level_array = np.array([[0, 1, 2], [3, 4]], dtype=object)
        plats_l = set(range(len(platforms_level_array)))

        # Set of platforms that form trailer and truck
        platforms_truck_trailer_array = np.array([[0, 1, 2, 3, 4]], dtype=object)
        plats_t = set(range(len(platforms_truck_trailer_array)))

        # Create decision variables
        x = pulp.LpVariable.dicts('x', ((p, v) for p in plats for v in vecs), cat='Binary')

        # Create the problem instance
        prob = pulp.LpProblem("ACL", pulp.LpMaximize)

        # Objective function
        # Maximize number of vehicles on the truck
        prob += pulp.lpSum(x[p, v] for p in plats for v in vecs)

        # Constraints
        # (1) Every vehicle can only be assigned to a single platform
        for p in plats:
            prob += pulp.lpSum(x[p, v] for v in vecs) <= 1

        # (2) Every platform can only hold a single vehicle
        for v in vecs:
            prob += pulp.lpSum(x[p, v] for p in plats) <= 1

        # (3) Weight limit for every platform
        for p in plats:
            for v in vecs:
                prob += weight_list[v] * x[p, v] <= wp[p]

        # (4) Weight constraint for every level
        for p_l in plats_l:
            prob += pulp.lpSum(weight_list[v] * x[p, v] for p in platforms_level_array[p_l] for v in vecs) <= \
                wl[p_l]

        # (5) Weight constraint for truck and trailer
        for t in plats_t:
            prob += pulp.lpSum(
                weight_list[v] * x[p, v] for p in platforms_truck_trailer_array[t] for v in vecs) <= wt[t]

        self.application = prob

    def _generate_small_model(self, df: any, vehicles: list) -> None:
        """
        Generate the problem model for the Small configuration

        :param df: Datafile
        :param vehicles: List of vehicle types
        """
        # Parameters for the small model (2 levels with 3 and 2 platforms each)

        # Length parameters
        # Level 1 (Truck up), 2 (Truck down)
        lmax_l = [97, 79]
        # Height parameters
        # Considers base truck height and height distance between vehicles (~10cm)
        hmax_truck = [34, 34, 33, 36, 32, 36]
        # Weight parameters
        # Max. total weight on truck / trailer
        wt = [100]
        # Max. weight on the two levels
        wl = [65, 50]
        # Max. weights on platforms p, if not angled
        wp = [23, 23, 23, 26, 17]
        # Max. weight on p, if sp is used
        wsp = [28, 28, 28]

        _, length_list, height_list, weight_list = self._get_vehicle_params(df, vehicles)

        # Set of available cars
        vecs = set(range(len(vehicles)))
        # Set of available platforms
        platforms_array = np.array([0, 1, 2, 3, 4])
        plats = set(range(len(platforms_array)))

        # Set of possible split platforms
        split_platforms_array = np.array([[0, 1], [1, 2], [3, 4]], dtype=object)
        plats_sp = set(range(len(split_platforms_array)))

        # Set of platforms that have a limitation on allowed length and weight because they are on the same level
        platforms_level_array = np.array([[0, 1, 2], [3, 4]], dtype=object)
        plats_l = set(range(len(platforms_level_array)))

        # Set of platforms that form trailer and truck
        platforms_truck_trailer_array = np.array([[0, 1, 2, 3, 4]], dtype=object)
        plats_t = set(range(len(platforms_truck_trailer_array)))

        # Set of platforms that have a limitation on allowed height
        platforms_height_array_truck = np.array([[0, 3], [1, 3], [2, 3], [0, 4], [1, 4], [2, 4]], dtype=object)
        plats_h1 = set(range(len(platforms_height_array_truck)))

        # Create decision variables
        x = pulp.LpVariable.dicts('x', ((p, v) for p in plats for v in vecs), cat='Binary')
        # Usage of split platform
        sp = pulp.LpVariable.dicts('sp', (q for q in plats_sp), cat='Binary')
        # Auxiliary variable for linearization of quadratic constraints
        gamma = pulp.LpVariable.dicts('gamma', (p for p in plats_sp), cat='Binary')

        # Create the 'prob' variable to contain the problem data
        prob = pulp.LpProblem("ACL", pulp.LpMaximize)
        # Maximize number of vehicles on the truck
        prob += pulp.lpSum(x[p, v] for p in plats for v in vecs)

        # Constraints
        # (1) Every vehicle can only be assigned to a single platform
        for p in plats:
            prob += pulp.lpSum(x[p, v] for v in vecs) <= 1

        # (2) Every platform can only hold a single vehicle
        for v in vecs:
            prob += pulp.lpSum(x[p, v] for p in plats) <= 1

        # (3) If a split platform q in plats_sp is used, only one of its "sub platforms" can be used
        for q in plats_sp:
            prob += pulp.lpSum(x[p, v] for p in split_platforms_array[q] for v in vecs) \
                <= len(split_platforms_array[q]) * (1 - sp[q]) + sp[q]

        # (4) It is always only possible to use a single split-platform for any given p
        for q in plats_sp:
            for p in plats_sp:
                if p != q:
                    z = bool(set(split_platforms_array[q]) & set(split_platforms_array[p]))
                    if z is True:
                        prob += sp[q] + sp[p] <= 1

        # (5) Length constraint
        # Checks that vehicles v on platforms p that belong to level L are shorter than the maximum available length
        for L in plats_l:
            prob += (pulp.lpSum(x[p, v] * length_list[v] for p in platforms_level_array[L] for v in vecs)
                     <= lmax_l[L])

        # (6) Height constraints for truck and trailer, analogue to length constraints
        # Truck
        for h in plats_h1:
            prob += pulp.lpSum(x[p, v] * height_list[v] for p in platforms_height_array_truck[h] for v in vecs) \
                <= hmax_truck[h]

        # (7) Linearization constraint -> gamma == 1, if split platform is used
        for q in plats_sp:
            prob += pulp.lpSum(
                sp[q] + x[p, v] for p in self.intersectset(split_platforms_array[q], platforms_array)
                for v in vecs) >= 2 * gamma[q]

        # (8) Weight limit for every platform
        for p in plats:
            for v in vecs:
                prob += weight_list[v] * x[p, v] <= wp[p]

        # (9) If a split platform is used, weight limit == wsp, if not, then weight limit == wp
        for q in plats_sp:
            for p in split_platforms_array[q]:
                prob += pulp.lpSum(weight_list[v] * x[p, v] for v in vecs) <= gamma[q] * wsp[q] \
                    + (1 - gamma[q]) * wp[p]

        # (10) Weight constraint for every level
        for p_l in plats_l:
            prob += pulp.lpSum(weight_list[v] * x[p, v] for p in platforms_level_array[p_l] for v in vecs) <= \
                wl[p_l]

        # (11) Weight constraint for truck and trailer
        for p_t in plats_t:
            prob += pulp.lpSum(
                weight_list[v] * x[p, v] for p in platforms_truck_trailer_array[p_t] for v in vecs) <= wt[p_t]

        self.application = prob

    def _generate_full_model(self, df: any, vehicles: list) -> None:  # pylint: disable=R0915
        """
        Generate the problem model for the Full configuration.

        :param df: Datafile
        :param vehicles: List of vehicle types
        """
        # Horizontal Coefficients: Length reduction
        # 1 = forward, 0 = backward
        v_coef = np.array([[0.20, 0.15, 0.14, 0.19],
                           [0.22, 0.22, 0.22, 0.22],
                           [0.22, 0.13, 0.12, 0.17]])

        # Vertical Coefficients: Height increase
        h_coef = np.array([[0.40, 1, 1, 1],
                           [0.17, 0.22, 0.21, 0.22],
                           [0.17, 0.38, 0.32, 0.32]])

        # Length parameters
        # Level 1 (Truck up), 2 (Truck down), 3 (Trailer up), 4 (Trailer down)
        lmax_l = [97, 79, 97, 97]

        # Height parameters
        # Considers base truck height and height distance between vehicles (~10cm)
        hmax_truck = [34, 34, 33, 36, 32, 36]
        hmax_trailer = [36, 32, 32, 34]

        # Weight parameters
        # Max. total weight
        wmax = 180
        # Max. total weight on truck / trailer
        wt = [100, 100]
        # Max. weight on the four levels
        wl = [50, 60, 50, 90]
        # Max. weights on platforms p, if not angled
        wp = [23, 23, 23, 26, 17, 26, 26, 26, 23, 26]
        # Max. weights on p, angled (if possible: 1, 2, 4, 7, 8, 9):
        wpa = [20, 22, 17, 18, 19, 22]
        # Max. weight on p, if sp is used
        wsp = [28, 28, 28, 28, 28, 28]

        class_list, length_list, height_list, weight_list = self._get_vehicle_params(df, vehicles)

        # Set of available cars
        vecs = set(range(len(vehicles)))
        # Set of available platforms
        platforms_array = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        plats = set(range(len(platforms_array)))

        # Set of platforms that can be angled
        platforms_angled_array = [1, 2, 4, 7, 8]
        vp = [0, 1, 3, 8, 9]
        plats_a = set(range(len(platforms_angled_array)))

        # Set of possible split platforms
        split_platforms_array = np.array([[0, 1], [1, 2], [3, 4], [5, 6], [7, 8], [8, 9]], dtype=object)
        plats_sp = set(range(len(split_platforms_array)))

        # Set of platforms that have a limitation on allowed length and weight because they are on the same level
        platforms_level_array = np.array([[0, 1, 2], [3, 4], [5, 6], [7, 8, 9]], dtype=object)
        plats_l = set(range(len(platforms_level_array)))

        # Set of platforms that form trailer and truck
        platforms_truck_trailer_array = np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=object)
        plats_t = set(range(len(platforms_truck_trailer_array)))

        # Set of platforms that have a limitation on allowed height
        platforms_height_array_truck = np.array([[0, 3], [1, 3], [2, 3], [0, 4], [1, 4], [2, 4]], dtype=object)
        platforms_height_array_trailer = np.array([[5, 7], [5, 8], [6, 8], [6, 9]], dtype=object)

        plats_h1 = set(range(len(platforms_height_array_truck)))
        plats_h2 = set(range(len(platforms_height_array_trailer)))

        # Create decision variables
        # Vehicle v assigned to p
        x = pulp.LpVariable.dicts('x', ((p, v) for p in plats for v in vecs), cat='Binary')
        # Usage of split platform
        sp = pulp.LpVariable.dicts('sp', (q for q in plats_sp), cat='Binary')
        # Direction of vehicle on p
        d = pulp.LpVariable.dicts('d', (p for p in plats), cat='Binary')
        # State of platform p in PA - angled == 1, not angled == 0
        a_p = pulp.LpVariable.dicts('a_p', (p for p in plats_a), cat='Binary')

        # Auxiliary variables for linearization of quadratic constraints
        y1 = pulp.LpVariable.dicts('y1', (p for p in plats_a), cat='Binary')
        y2 = pulp.LpVariable.dicts('y2', (p for p in plats_a), cat='Binary')
        y3 = pulp.LpVariable.dicts('y3', (p for p in plats_a), cat='Binary')
        y4 = pulp.LpVariable.dicts('y4', (p for p in plats_a), cat='Binary')
        ay1 = pulp.LpVariable.dicts('ay1', (p for p in plats_a), cat='Binary')
        ay2 = pulp.LpVariable.dicts('ay2', (p for p in plats_a), cat='Binary')
        ay3 = pulp.LpVariable.dicts('ay3', (p for p in plats_a), cat='Binary')
        ay4 = pulp.LpVariable.dicts('ay4', (p for p in plats_a), cat='Binary')
        # Weight for split-platforms
        gamma = pulp.LpVariable.dicts('gamma', (p for p in plats_sp), cat='Binary')

        # Create the 'prob' variable to contain the problem data
        prob = pulp.LpProblem("ACL", pulp.LpMaximize)
        # Maximize number of vehicles on the truck
        prob += pulp.lpSum(x[p, v] for p in plats for v in vecs)

        # Assignment constraints
        # (1) Every vehicle can only be assigned to a single platform
        for p in plats:
            prob += pulp.lpSum(x[p, v] for v in vecs) <= 1
        # (2) Every platform can only hold a single vehicle
        for v in vecs:
            prob += pulp.lpSum(x[p, v] for p in plats) <= 1

        # (3) If a split platform q in plats_sp is used, only one of its "sub platforms" can be used
        for q in plats_sp:
            prob += pulp.lpSum(x[p, v] for p in split_platforms_array[q] for v in vecs) \
                <= len(split_platforms_array[q]) * (1 - sp[q]) + sp[q]

        # (3.1) It is always only possible to use a single split-platform for any given p
        for q in plats_sp:
            for p in plats_sp:
                if p != q:
                    z = bool(set(split_platforms_array[q]) & set(split_platforms_array[p]))
                    if z is True:
                        prob += sp[q] + sp[p] <= 1

        # (3.2) It is not allowed to angle platforms next to empty platforms
        for i, p in enumerate(platforms_angled_array):
            prob += pulp.lpSum(x[p, v] + x[vp[i], v] for v in vecs) >= 2 * a_p[i]

        # Linearization constraints
        # Linearization of d_p and d_v(p) -> orientations of two neighboring cars
        for p in platforms_angled_array:
            z = platforms_angled_array.index(p)
            v_p = vp[z]
            prob += (1 - d[p]) + d[v_p] >= 2 * y1[z]
            prob += d[p] + d[v_p] >= 2 * y2[z]
            prob += (1 - d[p]) + (1 - d[v_p]) >= 2 * y3[z]
            prob += d[p] + (1 - d[v_p]) >= 2 * y4[z]

        # Linearization of a_p with y1 - y4 -> linear combination of angle and orientations
        for p in platforms_angled_array:
            z = platforms_angled_array.index(p)
            prob += a_p[z] + y1[z] >= 2 * ay1[z]
            prob += a_p[z] + y2[z] >= 2 * ay2[z]
            prob += a_p[z] + y3[z] >= 2 * ay3[z]
            prob += a_p[z] + y4[z] >= 2 * ay4[z]

        # Linearization of x * ay -> linear combination of assignment and orientation/angle
        xay1 = pulp.LpVariable.dicts('xay1', ((p, v) for p in plats_a for v in vecs), cat='Binary')
        xay2 = pulp.LpVariable.dicts('xay2', ((p, v) for p in plats_a for v in vecs), cat='Binary')
        xay3 = pulp.LpVariable.dicts('xay3', ((p, v) for p in plats_a for v in vecs), cat='Binary')
        xay4 = pulp.LpVariable.dicts('xay4', ((p, v) for p in plats_a for v in vecs), cat='Binary')
        for p in platforms_angled_array:
            z = platforms_angled_array.index(p)
            for v in vecs:
                prob += ay1[z] + x[z, v] >= 2 * xay1[z, v]
                prob += ay2[z] + x[z, v] >= 2 * xay2[z, v]
                prob += ay3[z] + x[z, v] >= 2 * xay3[z, v]
                prob += ay4[z] + x[z, v] >= 2 * xay4[z, v]

        # Making sure always only 1 case applies
        for p in platforms_angled_array:
            z = platforms_angled_array.index(p)
            prob += ay1[z] + ay2[z] + ay3[z] + ay4[z] <= 1
            prob += y1[z] + y2[z] + y3[z] + y4[z] <= 1

        # (4) Length constraint
        # Checks that vehicles v on platforms p that belong to level L are shorter than the maximum available length
        # The length of the vehicles depends on whether they are angled or not and which vehicle is standing on
        # platform v(p)
        for L in plats_l:
            prob += pulp.lpSum(x[p, v] * length_list[v]
                               - xay1[platforms_angled_array.index(p), v] *
                               int(v_coef[class_list[v]][0] * length_list[v])
                               - xay2[platforms_angled_array.index(p), v] *
                               int(v_coef[class_list[v]][1] * length_list[v])
                               - xay3[platforms_angled_array.index(p), v] *
                               int(v_coef[class_list[v]][2] * length_list[v])
                               - xay4[platforms_angled_array.index(p), v]
                               * int(v_coef[class_list[v]][3] * length_list[v])
                               for p in self.intersectset(platforms_angled_array, platforms_level_array[L])
                               for v in vecs) \
                + pulp.lpSum(x[p, v] * length_list[v]
                             for p in self.diffset(platforms_level_array[L], platforms_angled_array)
                             for v in vecs) \
                <= lmax_l[L]

        # (5) Platforms can not be angled, if they are part of a split platform
        for q in plats_sp:
            prob += pulp.lpSum(a_p[platforms_angled_array.index(p)]
                               for p in self.intersectset(platforms_angled_array, split_platforms_array[q])) \
                <= len(split_platforms_array[q]) * (1 - sp[q])

        # (6) Weight constraint if split platform is used, gamma == 1
        for q in plats_sp:
            prob += pulp.lpSum(sp[q] + x[p, v] for p in self.intersectset(split_platforms_array[q], platforms_array)
                               for v in vecs) >= 2 * gamma[q]

        # If split platform is used, weight limit == wsp, if not, then weight limit == wp
        for q in plats_sp:
            for p in split_platforms_array[q]:
                prob += (pulp.lpSum(weight_list[v] * x[p, v] for v in vecs) <= gamma[q] * wsp[q] + (1 - gamma[q]) *
                         wp[p])

        # (7) If a platform that can be angled is angled, weight limit == wpa
        # Need another linearization for that:
        apx = pulp.LpVariable.dicts('apx', ((p, v) for p in plats_a for v in vecs), cat='Binary')
        for p in platforms_angled_array:
            z = platforms_angled_array.index(p)
            for v in vecs:
                prob += a_p[z] + x[z, v] >= 2 * apx[z, v]

        for p in platforms_angled_array:
            prob += pulp.lpSum(weight_list[v] * apx[platforms_angled_array.index(p), v] for v in vecs) \
                <= wpa[platforms_angled_array.index(p)]

        # (8) Weight constraint for every level
        for p_l in plats_l:
            prob += (pulp.lpSum(weight_list[v] * x[p, v] for p in platforms_level_array[p_l] for v in vecs) <=
                     wl[p_l])

        # (9) Weight constraint for truck and trailer
        for p_t in plats_t:
            prob += (pulp.lpSum(weight_list[v] * x[p, v] for p in platforms_truck_trailer_array[p_t] for v in vecs)
                     <= wt[p_t])

        # (10) Weight constraint for entire auto carrier
        prob += pulp.lpSum(weight_list[v] * x[p, v] for p in plats for v in vecs) <= wmax

        # (11) Height constraints for truck and trailer, analogue to length constraints
        # Truck
        for h in plats_h1:
            prob += pulp.lpSum(x[p, v] * height_list[v]
                               - xay1[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][0] * height_list[v])
                               - xay2[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][1] * height_list[v])
                               - xay3[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][2] * height_list[v])
                               - xay4[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][3] * height_list[v])
                               for p in self.intersectset(platforms_angled_array, platforms_height_array_truck[h])
                               for v in vecs) \
                + pulp.lpSum(x[p, v] * height_list[v]
                             for p in self.diffset(platforms_height_array_truck[h], platforms_angled_array)
                             for v in vecs) \
                <= hmax_truck[h]
        # Trailer
        for h in plats_h2:
            prob += pulp.lpSum(x[p, v] * height_list[v]
                               - xay1[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][0] * height_list[v])
                               - xay2[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][1] * height_list[v])
                               - xay3[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][2] * height_list[v])
                               - xay4[platforms_angled_array.index(p), v] *
                               int(h_coef[class_list[v]][3] * height_list[v])
                               for p in self.intersectset(platforms_angled_array, platforms_height_array_trailer[h])
                               for v in vecs) \
                + pulp.lpSum(x[p, v] * height_list[v]
                             for p in self.diffset(platforms_height_array_trailer[h], platforms_angled_array)
                             for v in vecs) \
                <= hmax_trailer[h]

        self.application = prob

    def _get_vehicle_params(self, df: any, vehicles: list) -> tuple[list, list, list, list]:
        """
        Extract vehicle parameters for the problem formulation

        :param df: Dataframe containing vehicle data
        :param vehicles: List of vehicle types to consider
        :return: Lists containing class, length, height, and weight of vehicles
        """
        class_list = [0] * (len(vehicles))
        length_list = [0] * (len(vehicles))
        height_list = [0] * (len(vehicles))
        weight_list = [0] * (len(vehicles))

        for i in set(range(len(vehicles))):
            df_new = df.loc[df['Type'] == vehicles[i]]
            class_list[i] = int(df_new["Class"].iloc[0])
            length_list[i] = int(df_new["Length"].iloc[0])
            height_list[i] = int(df_new["Height"].iloc[0])
            weight_list[i] = int(df_new["Weight"].iloc[0])

        return class_list, length_list, height_list, weight_list

    def validate(self, solution: any) -> tuple[bool, float]:
        """
        Checks if the solution is a valid solution.
:
        :param solution: Proposed solution
        :return: Tuple containing a boolean indicating if the solution is valid
                 and the time it took to validate the solution
        """
        start = start_time_measurement()
        status = solution.get("status")
        is_valid = status == "Optimal"
        return is_valid, end_time_measurement(start)

    def get_solution_quality_unit(self) -> str:
        """
        Provides the unit of measure for solution quality.

        :return: The unit of measure fro solution quality
        """
        return "Number of loaded vehicles"

    def evaluate(self, solution: any) -> tuple[float, float]:
        """
        Checks how good the solution is.

        :param solution: Provided solution
        :return: Tuple containing the objective value and the time it took to evaluate the solution
        """
        start = start_time_measurement()
        objective_value = solution.get("obj_value", 0)
        logging.info("Loading successful!")
        logging.info(f"{objective_value} cars will fit on the auto carrier.")

        variables = solution.get("variables", {})
        assignments = [key for key in variables if variables[key] > 0]

        logging.info(f"vehicle-to-platform assignments (platform, vehicle): {assignments}")
        return objective_value, end_time_measurement(start)

    def save(self, path: str, iter_count: int) -> None:
        """
        Saves the problem instance to a JSON file.

        :param path: Directory path where the instance should be saved
        :param iter_count: Iteration count (unused)
        """
        # Convert our problem instance from Dict to an LP problem and then to json
        _, problem_instance = pulp.LpProblem.from_dict(self.application)
        # Save problem instance to JSON
        problem_instance.to_json(f"{path}/ACL_instance.json")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import product
from abc import ABC, abstractmethod

import numpy as np

from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class Transformation(Core, ABC):
    """
    The task of the transformation module is to translate data and problem
    specification of the application into preprocessed format.
    """

    def __init__(self, name):
        """
        Constructor method.
        """
        super().__init__()
        self.transformation_name = name

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "numpy", "version": "1.26.4"}]

    def preprocess(self, input_data: dict, config: dict, **kwargs: dict) -> tuple[dict, float]:
        """
        In this module, the preprocessing step is transforming the data to the correct target format.

        :param input_data: Collected information of the benchmarking process
        :param config: Config specifying the parameters of the transformation
        :param kwargs: Additional optional arguments
        :return: Tuple with transformed problem and the time it took to map it
        """
        start = start_time_measurement()
        output = self.transform(input_data, config)

        return output, end_time_measurement(start)

    def postprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[dict, float]:
        """
        Does the reverse transformation.

        :param input_data: Dictionary containing information of previously executed modules
        :param config: Dictionary containing additional information
        :param kwargs: Dictionary containing additional information
        :return: Tuple with the dictionary and the time the postprocessing took
        """
        start = start_time_measurement()
        output = self.reverse_transform(input_data)
        output["Transformation"] = True
        if "inference" in input_data:
            output["inference"] = input_data["inference"]

        return output, end_time_measurement(start)

    @abstractmethod
    def transform(self, input_data: dict, config: dict) -> dict:
        """
        Helps to ensure that the model can effectively learn the underlying
        patterns and structure of the data, and produce high-quality outputs.

        :param input_data: Input data for transformation
        :param config: Configuration parameters for the transformation
        :return: Transformed data.
        """
        return input_data

    def reverse_transform(self, input_data: dict) -> dict:
        """
        Transforms the solution back to the original problem.
        This might not be necessary in all cases, so the default is to return the original solution.
        This might be needed to convert the solution to a representation needed for validation and evaluation.

        :param input_data: The input data to be transformed
        :return: Transformed data
        """
        return input_data

    @staticmethod
    def compute_discretization(n_qubits: int, n_registered: int) -> np.ndarray:
        """
        Compute discretization for the grid.

        :param n_qubits: Total number of qubits
        :param n_registered: Number of qubits to be registered
        :return: Discretization data
        """
        n = 2 ** (n_qubits // n_registered)
        n_bins = n ** n_registered
        bin_data = np.empty((n_bins, n_registered + 1), dtype=float)

        for k, coords in enumerate(product(range(n), repeat=n_registered)):
            normalized_coords = np.array(coords) / n + 0.5 / n
            bin_data[k] = np.concatenate(([k], normalized_coords))

        return bin_data

    @staticmethod
    def compute_discretization_efficient(n_qubits: int, n_registers: int) -> np.ndarray:
        """
        Compute grid discretization.

        :param n_qubits: Total number of qubits
        :param n_registers: Number of qubits to be registered
        :return: Discretization data
        """
        n = 2 ** (n_qubits // n_registers)
        n_bins = n ** n_registers

        # Create an array of all possible coordinate combinations
        coords = np.array(list(product(range(n), repeat=n_registers)))

        # Calculate normalized_coords for all combinations
        normalized_coords = (coords.astype(float) + 0.5) / n

        # Create bin_data by concatenating normalized_coords with row indices
        bin_data = np.hstack((np.arange(n_bins).reshape(-1, 1), normalized_coords))

        return bin_data

    @staticmethod
    def generate_samples(results: np.ndarray, bin_data: np.ndarray, n_registers: int, noisy: bool = True) -> np.ndarray:
        """
        Generate samples based on measurement results and the grid bins.

        :param results: Results of measurements
        :param bin_data: Binned data
        :param n_registers: Number of registers
        :param noisy: Flag indicating whether to add noise
        :return: Generated samples
        """
        n_shots = np.sum(results)
        width = 1 / len(bin_data) ** (1 / n_registers)
        points = (
            0.5 * width * np.random.uniform(low=-1, high=1, size=(n_shots, n_registers))
            if noisy
            else np.zeros((n_shots, n_registers))
        )

        position = 0
        for idx, value in enumerate(results):
            bin_coords = bin_data[idx, 1:]
            points[position: position + value, :] += np.tile(bin_coords, (value, 1))
            position += value

        return points.astype(np.float32)

    @staticmethod
    def generate_samples_efficient(results, bin_data: np.ndarray, n_registers: int, noisy: bool = True) -> np.ndarray:
        """
        Generate samples efficiently using numpy arrays based on measurement results and the grid bins.

        :param results: Results of measurements
        :param bin_data: Binned data
        :param n_registers: Number of registers
        :param noisy: Flag indicating whether to add noise
        :return: Generated samples
        """
        n_shots = np.sum(results)
        width = 1 / len(bin_data) ** (1 / n_registers)

        # Generate random noise or zeros
        noise = (
            0.5 * width * np.random.uniform(low=-1, high=1, size=(n_shots, n_registers))
            if noisy
            else np.zeros((n_shots, n_registers))
        )

        # Create an array of bin_coords for each result, then stack them vertically
        bin_coords = bin_data[:, 1:]
        expanded_bin_coords = np.repeat(bin_coords, results, axis=0)

        # Reshape expanded_bin_coords to match the shape of noise
        expanded_bin_coords = expanded_bin_coords.reshape(n_shots, n_registers)

        # Add noise to the expanded_bin_coords
        points = expanded_bin_coords + noise

        return points.astype(np.float32)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict

import networkx as nx

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class ReverseGreedyClassicalPVC(Solver):
    """
    Classical Reverse Greedy Solver for the PVC problem. We take the worst choice at each step.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Local"]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """
        if option == "Local":
            from quark.modules.devices.Local import Local  # pylint: disable=C0415
            return Local()
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "networkx", "version": "3.4.2"}]

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this solver has no configurable settings.

        :return: Empty dict
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def run(self, mapped_problem: nx.Graph, device_wrapper: any, config: Config, **kwargs: dict) \
            -> tuple[dict, float, dict]:
        """
        Solve the PVC graph in a greedy fashion. We take the worst choice at each step.

        :param mapped_problem: Graph representing a PVC problem
        :param device_wrapper: Local device
        :param config: Empty dict
        :param kwargs: No additionally settings needed
        :return: Solution, the time it took to compute it and optional additional information
        """
        # Need to deep copy since we are modifying the graph in this function.
        # Else the next repetition would work with a different graph
        mapped_problem = mapped_problem.copy()
        start = start_time_measurement()

        # We always start at the base node
        current_node = ((0, 0), 1, 1)
        idx = 1

        tour = {current_node + (0,): 1}

        # Tour needs to cover all nodes, if there are 2 nodes left we can finish since these 2 nodes
        # belong to the same seam
        while len(mapped_problem.nodes) > 2:
            # Get the minimum neighbor edge from the current node
            # TODO This only works if the artificial high edge weights are exactly 100000
            next_node = max(
                (
                    x for x in mapped_problem.edges(current_node[0], data=True)
                    if x[2]['c_start'] == current_node[1]
                    and x[2]['t_start'] == current_node[2]
                    and x[2]['weight'] != 100000
                ),
                key=lambda x: x[2]['weight']
            )
            next_node = (next_node[1], next_node[2]["c_end"], next_node[2]["t_end"])

            # Make the step - add distance to cost, add the best node to tour,
            tour[next_node + (idx,)] = 1

            # Remove all node of that seam
            to_remove = [x for x in mapped_problem.nodes if x[0] == current_node[0][0]]
            for node in to_remove:
                mapped_problem.remove_node(node)

            current_node = next_node
            idx += 1

        # Tour needs to look like {((0, 0), 1, 1, 0): 1,((3, 1), 1, 0, 1): 1,((2, 1), 1, 1, 2): 1,((4, 4), 1, 1, 3): 1}
        # ((0, 0), 1, 1, 0): 1 = ((seam, node), config, tool, timestep): yes we visit this
        return tour, end_time_measurement(start), {}



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod

from quark.modules.devices.Device import Device


class Pulser(Device, ABC):
    """
    Abstract class to use the Pulser devices.
    """

    def __init__(self, device_name: str):
        """
        Constructor method.

        :param device_name: Name of the Pulser device.
        """
        super().__init__(device_name)
        self.device = None
        self.backend = None

    def get_backend(self) -> any:
        """
        Returns backend.

        :return: Instance of the backend class
        """
        return self.backend

    @abstractmethod
    def get_backend_config(self) -> any:
        """
        Returns backend configurations.

        :return: Instance of the backend config class
        """
        pass

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "pulser", "version": "1.1.1"}]



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict

import numpy as np
from dimod import qubo_to_ising

from quark.modules.applications.Mapping import Mapping, Core
from quark.modules.applications.optimization.SAT.mappings.ChoiQUBO import ChoiQUBO
from quark.utils import start_time_measurement, end_time_measurement


class ChoiIsing(Mapping):
    """
    Ising formulation for SAT problem using QUBO by Choi (1004.2226).
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["QAOA", "PennylaneQAOA"]
        self.problem = None
        self.qubo_mapping = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "dimod", "version": "0.12.18"},
            *ChoiQUBO.get_requirements()
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary with parameter options
        .. code-block:: python

            return {
                    "hard_reward": {
                        "values": [0.1, 0.5, 0.9, 0.99],
                        "description": "What Bh/A ratio do you want? (How strongly to enforce hard constraints)"
                    },
                    "soft_reward": {
                        "values": [0.1, 1, 2],
                        "description": "What Bh/Bs ratio do you want? This value is multiplied with the "
                                        "number of tests."
                    }
                }
        """
        return {
            "hard_reward": {
                "values": [0.1, 0.5, 0.9, 0.99],
                "description": "What Bh/A ratio do you want? (How strongly to enforce hard constraints)"
            },
            "soft_reward": {
                "values": [0.1, 1, 2],
                "description": "What Bh/Bs ratio do you want? This value is multiplied with the number of tests."
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             hard_reward: float
             soft_reward: float

        """
        hard_reward: float
        soft_reward: float

    def map(self, problem: any, config: Config) -> tuple[dict, float]:
        """
        Uses the ChoiQUBO formulation and converts it to an Ising.

        :param problem: SAT problem
        :param config: Dictionary with the mapping config
        :return: Dict with the ising, time it took to map it
        """
        start = start_time_measurement()
        self.problem = problem

        # call mapping function
        self.qubo_mapping = ChoiQUBO()
        q, _ = self.qubo_mapping.map(problem, config)
        t, j, _ = qubo_to_ising(q["Q"])

        # Convert Ising dict to matrix
        n = (len(problem[0]) + len(problem[1])) * 3
        t_vector = np.zeros(n, dtype=float)
        j_matrix = np.zeros((n, n), dtype=float)

        for key, value in t.items():
            t_vector[key] = value

        for key, value in j.items():
            j_matrix[key[0]][key[1]] = value

        return {"J": j_matrix, "t": t_vector}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the SAT class for validation/evaluation.

        :param solution: Dictionary containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()

        # convert raw solution into the right format to use reverse_map() of ChoiQUBO.py
        solution_dict = dict(enumerate(solution))

        # reverse map
        result, _ = self.qubo_mapping.reverse_map(solution_dict)

        return result, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "QAOA":
            from quark.modules.solvers.QAOA import QAOA  # pylint: disable=C0415
            return QAOA()
        if option == "PennylaneQAOA":
            from quark.modules.solvers.PennylaneQAOA import PennylaneQAOA  # pylint: disable=C0415
            return PennylaneQAOA()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TypedDict

import numpy as np

from qiskit.circuit.library import TwoLocal
from qiskit.primitives import Sampler, Estimator
from qiskit.quantum_info import SparsePauliOp, Statevector
from qiskit_optimization.applications import OptimizationApplication
from qiskit_algorithms.optimizers import POWELL, SPSA, COBYLA
from qiskit_algorithms.minimum_eigensolvers import VQE, QAOA, NumPyMinimumEigensolver

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class QiskitQAOA(Solver):
    """
    Qiskit QAOA.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["qasm_simulator", "qasm_simulator_gpu"]
        self.ry = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "qiskit", "version": "1.3.0"},
            {"name": "qiskit-optimization", "version": "0.6.1"},
            {"name": "numpy", "version": "1.26.4"},
            {"name": "qiskit-algorithms", "version": "0.3.1"}
        ]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """
        if option in ["qasm_simulator", "qasm_simulator_gpu"]:
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass(option)
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this solver.

        :return: Dictionary of configurable settings
        .. code-block:: python

                    return {
                            "shots": {  # number measurements to make on circuit
                                "values": list(range(10, 500, 30)),
                                "description": "How many shots do you need?"
                            },
                            "iterations": {  # number measurements to make on circuit
                                "values": [1, 5, 10, 20, 50, 75],
                                "description": "How many iterations do you need? Warning: When using\
                                the IBM Eagle Device you should only choose a lower number of\
                                iterations, since a high number would lead to a waiting time that\
                                could take up to multiple days!"
                            },
                            "depth": {
                                "values": [2, 3, 4, 5, 10, 20],
                                "description": "How many layers for QAOA (Parameter: p) do you want?"
                            },
                            "method": {
                                "values": ["classic", "vqe", "qaoa"],
                                "description": "Which Qiskit solver should be used?"
                            },
                            "optimizer": {
                                "values": ["POWELL", "SPSA", "COBYLA"],
                                "description": "Which Qiskit solver should be used? Warning: When\
                                using the IBM Eagle Device you should not use the SPSA optimizer,\
                                since it is not suited for only one evaluation!"
                            }
                        }

        """
        return {
            "shots": {  # number measurements to make on circuit
                "values": list(range(10, 500, 30)),
                "description": "How many shots do you need?"
            },
            "iterations": {  # number measurements to make on circuit
                "values": [1, 5, 10, 20, 50, 75],
                "description": "How many iterations do you need? Warning: When using the IBM Eagle device you should "
                               "only choose a low number of iterations (long computation times)."
            },
            "depth": {
                "values": [2, 3, 4, 5, 10, 20],
                "description": "How many layers for QAOA (parameter: p) do you want?"
            },
            "method": {
                "values": ["classic", "vqe", "qaoa"],
                "description": "Which Qiskit solver should be used?"
            },
            "optimizer": {
                "values": ["POWELL", "SPSA", "COBYLA"],
                "description": "Which Qiskit solver should be used? Warning: When using the IBM Eagle device you should"
                               " not use the SPSA optimizer for a low number of iterations!"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            shots: int
            depth: int
            iterations: int
            layers: int
            method: str
            optimizer: str

        """
        shots: int
        depth: int
        iterations: int
        layers: int
        method: str
        optimizer: str

    @staticmethod
    def normalize_data(data: any, scale: float = 1.0) -> any:
        """
        Not used currently, as I just scale the coefficients in the qaoa_operators_from_ising.

        :param data: Data to normalize
        :param scale: Scaling factor
        :return: Normalized data
        """
        return scale * data / np.max(np.abs(data))

    def run(self, mapped_problem: any, device_wrapper: any, config: Config, **kwargs: dict) -> tuple[any, float, dict]:
        """
        Run Qiskit QAOA algorithm on Ising.

        :param mapped_problem: Dictionary with the keys 'J' and 't'
        :param device_wrapper: Instance of device
        :param config: Config object for the solver
        :param kwargs: No additionally settings needed
        :return: Solution, the time it took to compute it and optional additional information
        """
        J = mapped_problem['J']
        t = mapped_problem['t']
        start = start_time_measurement()
        ising_op = self._get_pauli_op((t, J))

        if config["method"] == "classic":
            algorithm = NumPyMinimumEigensolver()
        else:
            optimizer = None
            if config["optimizer"] == "COBYLA":
                optimizer = COBYLA(maxiter=config["iterations"])
            elif config["optimizer"] == "POWELL":
                optimizer = POWELL(maxiter=config["iterations"], maxfev=config["iterations"] if
                                   device_wrapper.device == 'ibm_eagle' else None)
            elif config["optimizer"] == "SPSA":
                optimizer = SPSA(maxiter=config["iterations"])
            if config["method"] == "vqe":
                self.ry = TwoLocal(ising_op.num_qubits, "ry", "cz", reps=config["depth"], entanglement="full")
                estimator = Estimator()
                algorithm = VQE(ansatz=self.ry, optimizer=optimizer, estimator=estimator)
            elif config["method"] == "qaoa":
                sampler = Sampler()
                algorithm = QAOA(reps=config["depth"], optimizer=optimizer, sampler=sampler)
            else:
                logging.warning("No method selected in QiskitQAOA. Continue with NumPyMinimumEigensolver.")
                algorithm = NumPyMinimumEigensolver()

        # Run actual optimization algorithm
        try:
            result = algorithm.compute_minimum_eigenvalue(ising_op)
        except ValueError as e:
            logging.error(f"The following ValueError occurred in module QiskitQAOA: {e}")
            logging.error("The benchmarking run terminates with exception.")
            raise Exception("Please refer to the logged error message.") from e

        best_bitstring = self._get_best_solution(result)
        return best_bitstring, end_time_measurement(start), {}

    def _get_best_solution(self, result) -> any:
        """
        Gets the best solution from the result.

        :param result: Result from the quantum algorithm
        :return: Best bitstring solution
        """
        if self.ry is not None:
            if hasattr(result, "optimal_point"):
                para_dict = dict(zip(self.ry.parameters, result.optimal_point))
                unbound_para = set(self.ry.parameters) - set(para_dict.keys())
                for param in unbound_para:
                    para_dict[param] = 0.0
                eigvec = Statevector(self.ry.assign_parameters(para_dict))
            elif hasattr(result, "eigenstate"):
                eigvec = result.eigenstate
            else:
                raise AttributeError("The result object does not have 'optimal_point' or 'eigenstate' attributes.")
        else:
            if hasattr(result, "eigenstate"):
                eigvec = result.eigenstate
            else:
                raise AttributeError("The result object does not have 'eigenstate'.")

        best_bitstring = OptimizationApplication.sample_most_likely(eigvec)
        return best_bitstring

    @staticmethod
    def _get_pauli_op(ising: tuple[np.ndarray, np.ndarray]) -> SparsePauliOp:
        """
        Creates a Pauli operator from the given Ising model representation.

        :param ising: Tuple with linear and quandratic terms
        .return: SparsePauliOp representing the Ising model
        """
        pauli_list = []
        number_qubits = len(ising[0])

        # Linear terms
        it = np.nditer(ising[0], flags=['multi_index'])
        for x in it:
            logging.debug(f"{x},{it.multi_index}")
            key = it.multi_index[0]
            pauli_str = "I" * number_qubits
            pauli_str_list = list(pauli_str)
            pauli_str_list[key] = "Z"
            pauli_str = "".join(pauli_str_list)
            pauli_list.append((pauli_str, complex(x)))

        # J Part with 2-Qubit interactions
        it = np.nditer(ising[1], flags=['multi_index'])
        for x in it:
            logging.debug(f"{x},{it.multi_index}")
            idx1 = it.multi_index[0]
            idx2 = it.multi_index[1]
            pauli_str = "I" * number_qubits
            pauli_str_list = list(pauli_str)
            pauli_str_list[idx1] = "Z"
            pauli_str_list[idx2] = "Z"
            pauli_str = "".join(pauli_str_list)
            pauli_list.append((pauli_str, complex(x)))

        ising_op = SparsePauliOp.from_list(pauli_list)
        return ising_op



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from quark.modules.devices.Device import Device


class HelperClass(Device):
    """
    Some solvers like Pennylane, only needs strings for setting up the device and not a standalone class.

    TODO: Maybe refactor this once we think of a better structure for this
    """

    def __init__(self, device_name: str):
        """
        Constructor method.

        :param device_name: The name of the device
        """
        super().__init__(device_name=device_name)
        self.device = device_name
        self.submodule_options = []

    def get_parameter_options(self) -> dict:
        """
        Returns empty dictionary as this solver has no configurable settings.

        :return: Empty dict
        """
        return {}

    def get_default_submodule(self, option: str) -> None:
        """
        Raises ValueError as this module has no submodules.

        :param option: Option name
        :raises ValueError: If called, since this module has no submodules
        """
        raise ValueError("This module has no submodules.")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import json
import os
import time
from pathlib import Path
import inspect

import yaml
from packaging import version
import inquirer

from quark.modules.Core import Core
from quark.utils import _get_instance_with_sub_options, get_git_revision, checkbox


class Installer:
    """
    Installer class that can be used by the user to install certain QUARK modules and also return the required Python
    packages for the demanded modules.
    """

    def __init__(self):
        self.quark_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
        self.settings_dir = f"{self.quark_dir}/.settings"
        self.envs_dir = f"{self.settings_dir}/envs"
        self.python_version = "3.12.7"
        self.pip_version = "23.0"
        self.default_app_modules = [
            {"name": "PVC", "class": "PVC", "module": "modules.applications.optimization.PVC.PVC"},
            {"name": "SAT", "class": "SAT", "module": "modules.applications.optimization.SAT.SAT"},
            {"name": "TSP", "class": "TSP", "module": "modules.applications.optimization.TSP.TSP"},
            {"name": "ACL", "class": "ACL", "module": "modules.applications.optimization.ACL.ACL"},
            {"name": "MIS", "class": "MIS", "module": "modules.applications.optimization.MIS.MIS"},
            {"name": "SCP", "class": "SCP", "module": "modules.applications.optimization.SCP.SCP"},
            {"name": "GenerativeModeling", "class": "GenerativeModeling",
             "module": "modules.applications.qml.generative_modeling.GenerativeModeling"}
        ]

        self.core_requirements = [
            {"name": "seaborn", "version": "0.13.2"},
            {"name": "networkx", "version": "3.4.2"},
            {"name": "inquirer", "version": "3.4.0"},
            {"name": "packaging", "version": "24.2"},
            {"name": "pyyaml", "version": "6.0.2"},
            {"name": "typing-extensions", "version": "4.12.2"},
            {"name": "sphinx", "version": "8.1.3"},
            {"name": "sphinx-rtd-theme", "version": "3.0.2"},
        ]
        Path(self.envs_dir).mkdir(parents=True, exist_ok=True)

    def configure(self, env_name="default") -> None:
        """
        Configures a new QUARK environment or overwrites an existing one.

        :param env_name: Name of the env to configure
        """
        configured_envs = self.check_for_configs()

        if env_name in configured_envs:
            answer_continue = inquirer.prompt([
                inquirer.List("continue",
                              message=f"{env_name} found in the list of existing QUARK module environment, are you"
                              f" sure you want to overwrite it?",
                              choices=["Yes", "No"], )])["continue"]

            if answer_continue.lower() == "no":
                logging.info("Exiting configuration")
                return

        chosen_config_type = inquirer.prompt([
            inquirer.List("config",
                          message="Do you want to use the default configuration or a custom environment?",
                          choices=["Default", "Custom"])])["config"]
        logging.info(f"You chose {chosen_config_type}")

        module_db = self.get_module_db()

        if chosen_config_type == "Default":
            self.save_env(module_db, env_name)
        elif chosen_config_type == "Custom":
            module_db = self.start_query_user(module_db)
            self.save_env(module_db, env_name)

        requirements = self.collect_requirements(module_db["modules"])
        activate_requirements = checkbox("requirements", "Should we create an package file, if yes for "
                                                         "which package manager?",
                                         ["Conda", "PIP", "Print it here"])["requirements"]

        if "Conda" in activate_requirements:
            self.create_conda_file(requirements, env_name)
        if "PIP" in activate_requirements:
            self.create_req_file(requirements, env_name)
        if "Print it here" in activate_requirements:
            logging.info("Please install:")
            for p, v in requirements.items():
                logging.info(f"  -  {p}{': ' + v[0] if v else ''}")

        activate_answer = inquirer.prompt([
            inquirer.List("activate",
                          message="Do you want to activate the QUARK module environment?",
                          choices=["Yes", "No"])])["activate"]

        if activate_answer == "Yes":
            self.set_active_env(env_name)

    def check_for_configs(self) -> list:
        """
        Checks if QUARK is already configured and if yes, which environments.

        :return: Returns the configured QUARK envs in a list
        """
        return list(p.stem for p in Path(self.envs_dir).glob("*.json"))

    def set_active_env(self, name: str) -> None:
        """
        Sets the active env to active_env.json.

        :param name: Name of the env
        """
        self._check_if_env_exists(name)
        with open(f"{self.settings_dir}/active_env.json", "w") as jsonFile:
            json.dump({"name": name}, jsonFile, indent=2)

        logging.info(f"Set active QUARK module environment to {name}")

    def check_active_env(self) -> bool:
        """
        Checks if .settings/active_env.json exists.

        :return: True if active_env.json exists
        """
        return Path(f"{self.settings_dir}/active_env.json").is_file()

    def get_active_env(self) -> str:
        """
        Returns the current active environment.

        :return: Returns the name of the active env
        """
        # if not self.check_active_env():
        #     logging.warning("No active QUARK module environment found, using default")
        #     module_db = self.get_module_db()
        #     self.save_env(module_db, "default")
        #     self.set_active_env("default")

        # with open(f"{self.settings_dir}/active_env.json", "r") as filehandler:
        #     env = json.load(filehandler)
        #     return env["name"]

        return ""

    def get_env(self, name: str) -> list[dict]:
        """
        Loads the env from file and returns it.

        :param name: Name of the env
        :return: Returns the modules of the env
        """
        # file = f"{self.envs_dir}/{name}.json"
        # self._check_if_env_exists(name)

        # with open(file, "r") as filehandler:
        #     env = json.load(filehandler)
        #     logging.info(f"Getting {name} QUARK module environment")
        #     module_db_build_number = self.get_module_db_build_number()
        #     if env["build_number"] < module_db_build_number:
        #         logging.warning(f"You QUARK module env is based on an outdated build version of the module database "
        #                         f"(BUILD NUMBER {env['build_number']}). The current module database (BUILD NUMBER "
        #                         f"{module_db_build_number}) might bring new features. You should think about "
        #                         f"updating your environment!")

        #     return env["modules"]
        return self.default_app_modules

    def _check_if_env_exists(self, name: str) -> str:
        """
        Checks if a given env exists, returns the location of the associated JSON file and raises an error otherwise.

        :param name: Name of the env
        :return: Returns location of the JSON file associated with the env if it exists
        """
        file = f"{self.envs_dir}/{name}.json"
        if not Path(file).is_file():
            raise ValueError(f"QUARK environment {name} could not be found!")
        return file

    def save_env(self, env: dict, name: str) -> None:
        """
        Saves a created env to a file with the name of choice.

        :param env: Env which should be saved
        :param name: Name of the env
        """
        with open(f"{self.envs_dir}/{name}.json", "w") as jsonFile:
            json.dump(env, jsonFile, indent=2)

        logging.info(f"Saved {name} QUARK module environment.")

    def start_query_user(self, module_db: dict) -> dict:
        """
        Queries the user which applications and submodules to include.

        :param module_db: module_db file
        :return: Returns the module_db with selected (sub)modules
        """
        answer_apps = checkbox("apps", "Which application would you like to include?",
                               [m["name"] for m in module_db["modules"]])["apps"]

        module_db["modules"] = [x for x in module_db["modules"] if x["name"] in answer_apps]

        for idx, entry in enumerate(module_db["modules"]):
            self.query_user(module_db["modules"][idx], entry["name"])

        return module_db

    def query_user(self, submodules: dict, name: str) -> None:
        """
        Queries the user which submodules to include

        :param submodules: Submodules for the module
        :param name: Name of the module
        """
        if submodules["submodules"]:
            answer_submodules = \
                checkbox("submodules", f"Which submodule would you like to include for {name}?",
                         [m["name"] for m in submodules["submodules"]])["submodules"]

            submodules["submodules"] = [x for x in submodules["submodules"] if x["name"] in answer_submodules]
            for idx, entry in enumerate(submodules["submodules"]):
                self.query_user(submodules["submodules"][idx], entry["name"])

    def get_module_db(self) -> dict:
        """
        Returns the module database that contains all module possibilities.

        :return: Module Database
        """
        return {
            "build_number": 1,
            "build_date": time.strftime("%d-%m-%Y %H:%M:%S", time.gmtime()),
            "git_revision_number": 1,
            "modules": self.default_app_modules
        }

    def create_module_db(self) -> None:
        """
        Creates module database by automatically going through the available submodules for each module.
        """
        logging.info("Creating Module Database")

        # TODO Helper to skip constructor in Braket.py. Should probably be changed in the future!
        os.environ["SKIP_INIT"] = "True"

        module_db_modules: list[dict] = self.default_app_modules

        for idx, app in enumerate(module_db_modules):
            logging.info(f"Processing {app['name']}")
            app_instance = _get_instance_with_sub_options(module_db_modules, app["name"])
            module_db_modules[idx]["submodules"] = [
                Installer._create_module_db_helper(app_instance.get_submodule(sm), sm) for
                sm in app_instance.submodule_options]
            module_db_modules[idx]["requirements"] = app_instance.get_requirements()

        git_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", )
        git_revision_number, _ = get_git_revision(git_dir)
        module_db = {
            "build_number": self.get_module_db_build_number() + 1,
            "build_date": time.strftime("%d-%m-%Y %H:%M:%S", time.gmtime()),
            "git_revision_number": git_revision_number,
            "modules": module_db_modules
        }

        # loop over first level and add to file
        with open(f"{self.settings_dir}/module_db.json", "w") as jsonFile:
            json.dump(module_db, jsonFile, indent=2)

        requirements = self.collect_requirements(module_db_modules)
        self.create_req_file(requirements, "full", self.settings_dir)

        logging.info("Created module_db file")

    @staticmethod
    def _create_module_db_helper(module: Core, name: str) -> dict:
        """
        Recursive helper function for create_module_db.

        :param module: Module instance
        :param name: Name of the module
        :return: Module dict
        """
        return {
            "name": name,
            "class": module.__class__.__name__,
            "args": {k: v for k, v in module.__dict__.items() if k in
                     inspect.signature(module.__init__).parameters.keys()},
            "module": module.__module__,
            "requirements": module.get_requirements(),
            "submodules": [Installer._create_module_db_helper(module.get_submodule(sm), sm) for sm in
                           module.submodule_options]
        }

    def get_module_db_build_number(self) -> int:
        """
        Returns the build number of the module_db.

        :return: Returns the build number of the module_db if it exists, otherwise 0
        """
        if Path(f"{self.settings_dir}/module_db.json").is_file():
            module_db = self.get_module_db()
            return module_db["build_number"]
        else:
            return 0

    def collect_requirements(self, env: list[dict]) -> dict:
        """
        Collects requirements of the different modules in the given env file.

        :param env: Environment configuration
        :return: Collected requirements
        """
        requirements: list[dict] = self.core_requirements
        for app in env:
            requirements.extend(Installer._collect_requirements_helper(app))

        # Counts duplicates and checks if any version conflicts occur
        merged_requirements: dict = {}
        for req in requirements:
            if req["name"] in merged_requirements:
                # Checks if the specific version is already there, if so the req is skipped
                if merged_requirements[req["name"]] and ("version" in req and req["version"] not in
                                                         merged_requirements[req["name"]]):
                    merged_requirements[req["name"]].append(req["version"])

            else:
                # Sometimes there is no specific version required, then the "version" field is missing
                merged_requirements[req["name"]] = [req["version"]] if "version" in req else []

        for k, v in merged_requirements.items():
            if len(v) > 1:
                # If there are multiple different versions, the latest version is selected
                newest_version = sorted(v, key=lambda x: version.Version(x))[-1]  # pylint: disable=W0108
                merged_requirements[k] = [newest_version]
                logging.warning(f"Different version requirements for {k}: {v}. Will try to take the newer version "
                                f"{newest_version}, but this might cause problems at QUARK runtime!")

        return merged_requirements

    @staticmethod
    def _collect_requirements_helper(module: dict) -> list[dict]:
        """
        Helper function for collect_requirements_helper that recursively checks modules for requirements.

        :param module: Module dict
        :return: List of dicts with the requirements
        """
        requirements = module["requirements"]
        for submodule in module["submodules"]:
            requirements.extend(Installer._collect_requirements_helper(submodule))

        return requirements

    def create_conda_file(self, requirements: dict, name: str, directory: str = None) -> None:
        """
        Creates conda yaml file based on the requirements.

        :param requirements: Collected requirements
        :param name: Name of the conda env
        :param directory: Directory where the file should be saved. If None self.envs_dir will be taken.
        """
        if directory is None:
            directory = self.envs_dir
        conda_data = {
            "name": name,
            "channels": ["defaults"],
            "dependencies": [
                f"python={self.python_version}",
                f"pip={self.pip_version}",
                {"pip": [(f"{k}=={v[0]}" if v else k) for k, v in requirements.items()]}

            ]
        }
        with open(f"{directory}/conda_{name}.yml", "w") as filehandler:
            yaml.dump(conda_data, filehandler)

        logging.info("Saved conda env file, if you like to install it run:")
        logging.info(f"conda env create -f {directory}/conda_{name}.yml")

    def create_req_file(self, requirements: dict, name: str, directory: str = None) -> None:
        """
        Creates pip txt file based on the requirements.

        :param requirements: Collected requirements
        :param name: Name of the env
        :param directory: Directory where the file should be saved. If None self.envs_dir will be taken.
        """
        if directory is None:
            directory = self.envs_dir
        with open(f"{directory}/requirements_{name}.txt", "w") as filehandler:
            for k, v in requirements.items():
                filehandler.write(f"{k}=={v[0]}" if v else k)
                filehandler.write("\n")

        logging.info("Saved pip txt file, if you like to install it run:")
        logging.info(f"pip install -r {directory}/requirements_{name}.txt")

    def list_envs(self) -> None:
        """
        List all existing environments.
        """
        logging.info("Existing environments:")
        for env in self.check_for_configs():
            logging.info(f"  -  {env}")

    @staticmethod
    def show(env: list[dict]) -> None:
        """
        Visualize the env.

        :param env: Environment configuration
        """
        space = "    "
        branch = "|   "
        connector = "|-- "
        leaf = ">-- "

        def tree(modules: list[dict], prefix: str = ""):
            """
             A recursive function that generates a tree from the modules.
             This function is based on https://stackoverflow.com/a/59109706, but modified to the needs here.

            :param modules: Modules list
            :param prefix: Prefix for the indentation
            :return: Generator yielding formatted lines of the environment tree
            """
            # Modules in the middle/beginning get a |--, the final leaf >--
            pointers = [connector] * (len(modules) - 1) + [leaf]
            for pointer, module in zip(pointers, modules):
                yield prefix + pointer + module["name"]

                if module["submodules"]:
                    # If the module has any submodules
                    extension = branch if pointer == connector else space
                    # Check if we are at the end of the tree
                    yield from tree(module["submodules"], prefix=prefix + extension)

        logging.info("Content of the environment:")
        for module in tree(env):
            logging.info(module)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from quark.modules.Core import Core


class Solver(Core, ABC):
    """
    The solver is responsible for finding feasible and high-quality solutions of the formulated problem, i.e., of the
    defined objective function.
    """

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        The actual solving process is done here, using the device which is provided by the device submodule
        and the problem data provided by the parent module.

        :param input_data: Data passed to the run function of the solver
        :param config: Solver config
        :param kwargs: Optional keyword arguments
        :return: Output and time needed
        """
        output, elapsed_time, additional_metrics = self.run(self.preprocessed_input, input_data, config, **kwargs)
        self.metrics.add_metric_batch(additional_metrics)
        return output, elapsed_time

    @abstractmethod
    def run(self, mapped_problem: any, device_wrapper: any, config: any, **kwargs) -> tuple[any, float, dict]:
        """
        This function runs the solving algorithm on a mapped problem instance and returns a solution.

        :param mapped_problem: A representation of the problem that the solver can solve
        :param device_wrapper: A device the solver can leverage for the algorithm
        :param config: Settings for the solver such as hyperparameters
        :param kwargs: Optional additional settings
        :return: Solution, the time it took to compute it and some optional additional information
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from abc import ABC, abstractmethod
import logging

from quark.modules.applications.Application import Application
from quark.utils import start_time_measurement, end_time_measurement


class Optimization(Application, ABC):
    """
    Optimization Module for QUARK, is used by all Optimization applications.
    """

    @abstractmethod
    def validate(self, solution: any) -> tuple[bool, float]:
        """
        Checks if the solution is a valid solution.

        :param solution: Proposed solution
        :return: Bool value if solution is valid and the time it took to validate the solution
        """
        pass

    @abstractmethod
    def get_solution_quality_unit(self) -> str:
        """
        Returns the unit of the evaluation.

        :return: String with the unit
        """
        pass

    @abstractmethod
    def evaluate(self, solution: any) -> tuple[float, float]:
        """
        Checks how good the solution is.

        :param solution: Provided solution
        :return: Tuple with the evaluation and the time it took to create it
        """
        pass

    @abstractmethod
    def generate_problem(self, config: dict) -> any:
        """
        Creates a concrete problem and returns it.

        :param config: Configuration for problem creation
        :return: Generated problem
        """
        pass

    def process_solution(self, solution: any) -> tuple[any, float]:
        """
        Most of the time the solution has to be processed before it can be validated and evaluated.
        This might not be necessary in all cases, so the default is to return the original solution.

        :param solution: Proposed solution
        :return: Tuple with processed solution and the execution time to process it
        """
        return solution, 0.0

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        For optimization problems, we generate the actual problem instance in the preprocess function.

        :param input_data: Input data (usually not used in this method)
        :param config: Config for the problem creation
        :param kwargs: Optional additional arguments
        :return: Tuple with output and the preprocessing time
        """
        start = start_time_measurement()
        output = self.generate_problem(config)
        return output, end_time_measurement(start)

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        For optimization problems, we process the solution here, then validate and evaluate it.

        :param input_data: Data which should be evaluated for this optimization problem
        :param config: Config for the problem creation
        :param kwargs: Optional additional arguments
        :return: Tuple with results and the postprocessing time
        """
        processed_solution = None
        try:
            processed_solution, time_to_process_solution = self.process_solution(input_data)
            solution_validity, time_to_validation = self.validate(processed_solution)
        except Exception as e:
            logging.exception(f"Exception on processing the solution: {e}")
            solution_validity = False
            time_to_process_solution = None
            time_to_validation = None

        if solution_validity and (processed_solution is not None):
            solution_quality, time_to_evaluation = self.evaluate(processed_solution)
            self.visualize_solution(processed_solution, f"{kwargs["store_dir"]}/solution.pdf")
        else:
            solution_quality = None
            time_to_evaluation = None

        self.metrics.add_metric_batch({
            "application_score_value": solution_quality,
            "application_score_unit": self.get_solution_quality_unit(),
            "application_score_type": str(float),
            "processed_solution": processed_solution,
            "time_to_process_solution": time_to_process_solution,
            "time_to_validation": time_to_validation,
            "time_to_evaluation": time_to_evaluation
        })

        return solution_validity, sum(filter(None, [
            time_to_process_solution, time_to_validation, time_to_evaluation
        ]))

    def visualize_solution(self, processed_solution: any, path: str) -> None:
        """
        Creates visualizations of a processed and validated solution and writes them to disk.
        Override if applicable. Default is to do nothing.

        :param processed_solution: A solution that was already processed by :func:`process_solution`
        :param path: File path for the plot
        :returns: None
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging

import numpy as np

from pysat.formula import WCNF

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class RandomSAT(Solver):
    """
    Classic Random Solver for the SAT problem.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Local"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "python-sat", "version": "1.8.dev13"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_default_submodule(self, option: str) -> Core:
        if option == "Local":
            from quark.modules.devices.Local import Local  # pylint: disable=C0415
            return Local()
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this solver has no configurable settings.

        :return: Empty dict
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def run(self, mapped_problem: WCNF, device_wrapper: any, config: Config, **kwargs: dict) \
            -> tuple[list, float, dict]:
        """
        The given application is a problem instance from the pysat library.
        This generates a random solution to the problem.

        :param mapped_problem: The WCNF representation of the SAT problem
        :param device_wrapper: Local device
        :param config: Empty dict
        :param kwargs: No additionally settings needed
        :return: Solution, the time it took to compute it and optional additional information
        """
        logging.info(
            f"Got SAT problem with {mapped_problem.nv} variables, {len(mapped_problem.hard)} constraints and"
            f" {len(mapped_problem.soft)} tests."
        )

        start = start_time_measurement()
        sol = [(i + 1) * np.random.choice([-1, 1]) for i in range(mapped_problem.nv)]

        return sol, end_time_measurement(start), {}



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging

import torch
from torch.utils.data import DataLoader
from torch import nn
import torch.nn.functional as funct
from tensorboardX import SummaryWriter
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from quark.modules.applications.qml.generative_modeling.training.TrainingGenerative import TrainingGenerative, Core
from quark.utils_mpi import is_running_mpi, get_comm

matplotlib.use('Agg')
MPI = is_running_mpi()
comm = get_comm()


class QGAN(TrainingGenerative):  # pylint: disable=R0902
    """
    Class for QGAN
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("QGAN")

        self.beta_1 = None
        self.real_label = None
        self.fake_label = None
        self.n_qubits = None
        self.n_registers = None
        self.n_shots = None
        self.train_size = None
        self.execute_circuit = None
        self.device = None
        self.n_epochs = None
        self.batch_size = None
        self.learning_rate_generator = None
        self.n_bins = None
        self.n_states_range = None
        self.timing = None
        self.writer = None
        self.bins_train = None
        self.bins_train = None
        self.study_generalization = None
        self.generalization_metrics = None
        self.target = None
        self.n_params = None
        self.discriminator = None
        self.params = None
        self.generator = None
        self.accuracy = None
        self.criterion = None
        self.optimizer_discriminator = None
        self.real_labels = None
        self.fake_labels = None
        self.dataloader = None
        self.loss_func = None
        self.params = None
        self.discriminator_weights = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "torch", "version": "2.5.1"},
            {"name": "matplotlib", "version": "3.9.3"},
            {"name": "tensorboard", "version": "2.18.0"},
            {"name": "tensorboardX", "version": "2.6.2.2"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this circuit.

        :return: Configuration settings for QGAN
        .. code-block:: python
        return {
            "epochs": {
                "values": [2, 100, 200, 10000],
                "description": "How many epochs do you want?"
            },
            "batch_size": {
                "values": [10, 20, 100, 2000],
                "description": "What batch size do you want?"
            },
            "learning_rate_generator": {
                "values": [0.1, 0.2],
                "description": "What learning rate do you want to set for the generator?"
            },
            "learning_rate_discriminator": {
                "values": [0.1, 0.05],
                "description": "What learning rate do you want to set for the discriminator?"
            },
            "device": {
                "values": ["cpu", "gpu"],
                "description": "Where do you want to run the discriminator?"
            },
            "pretrained": {
                "values": [True, False],
                "description": "Do you want to use parameters of a pretrained model?"
            },
            "loss": {
                "values": ["KL", "NLL"],
                "description": "Which loss function do you want to use?"
            }
                    }
        """
        return {

            "epochs": {
                "values": [2, 100, 200, 10000],
                "description": "How many epochs do you want?"
            },
            "batch_size": {
                "values": [10, 20, 100, 2000],
                "description": "What batch size do you want?"
            },
            "learning_rate_generator": {
                "values": [0.1, 0.2],
                "description": "What learning rate do you want to set for the generator?"
            },
            "learning_rate_discriminator": {
                "values": [0.1, 0.05],
                "description": "What learnig rate do you want to set for the discriminator?"
            },
            "device": {
                "values": ["cpu", "gpu"],
                "description": "Where do you want to run the discriminator?"
            },
            "pretrained": {
                "values": [True, False],
                "description": "Do you want to use parameters of a pretrained model?"
            },
            "loss": {
                "values": ["KL", "NLL"],
                "description": "Which loss function do you want to use?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            epochs: int
            batch_size: int
            learning_rate_generator: int
            learning_rate_discriminator: int
            device: str
            loss: str

        """

        epochs: int
        batch_size: int
        learning_rate_generator: float
        learning_rate_discriminator: float
        device: str
        loss: str

    def get_default_submodule(self, option: str) -> Core:
        """
        Raises ValueError as this module has no submodules.

        :param option: Option name
        :raises ValueError: If called, since this module has no submodules
        """
        raise ValueError("This module has no submodules.")

    def setup_training(self, input_data: dict, config: dict) -> None:
        """
        Sets up the training configuration.

        :param input_data: dictionary with the variables from the circuit needed to start the training
        :param config: Configurations for the QGAN training.
        """
        self.beta_1 = 0.5
        self.real_label = 1.
        self.fake_label = 0.

        self.n_qubits = input_data['n_qubits']
        self.n_registers = input_data['n_registers']
        self.n_shots = input_data["n_shots"]
        self.train_size = input_data["train_size"]
        self.execute_circuit = input_data["execute_circuit"]

        self.device = config["device"]
        self.n_epochs = config["epochs"]
        self.batch_size = config["batch_size"]
        self.learning_rate_generator = config["learning_rate_generator"]

        n = 2 ** (self.n_qubits // self.n_registers)
        self.n_bins = n ** self.n_registers
        self.n_states_range = range(2 ** self.n_qubits)

        self.timing = self.Timing()
        self.writer = SummaryWriter(input_data["store_dir_iter"])

        self.bins_train = input_data["binary_train"]
        if input_data["dataset_name"] == "Cardinality_Constraint":
            new_size = 1000
            self.bins_train = np.repeat(self.bins_train, new_size, axis=0)

        self.study_generalization = "generalization_metrics" in list(input_data.keys())
        if self.study_generalization:
            self.generalization_metrics = input_data["generalization_metrics"]

        self.target = np.asarray(input_data["histogram_train"])
        self.target[self.target == 0] = 1e-8

        self.n_params = input_data["n_params"]

        self.discriminator = Discriminator(self.n_qubits).to(self.device)
        self.discriminator.apply(Discriminator.weights_init)

        self.params = np.random.rand(self.n_params) * np.pi
        self.generator = QuantumGenerator(self.n_qubits, self.execute_circuit, self.batch_size)

        self.accuracy = []
        self.criterion = torch.nn.BCELoss()
        self.optimizer_discriminator = torch.optim.Adam(
            self.discriminator.parameters(),
            lr=config["learning_rate_discriminator"],
            betas=(self.beta_1, 0.999)
        )

        self.real_labels = torch.full((self.batch_size,), 1.0, dtype=torch.float, device=self.device)
        self.fake_labels = torch.full((self.batch_size,), 0.0, dtype=torch.float, device=self.device)

        self.dataloader = DataLoader(self.bins_train, batch_size=self.batch_size, shuffle=True, drop_last=True)

        if config['loss'] == "KL":
            self.loss_func = self.kl_divergence
        elif config['loss'] == "NLL":
            self.loss_func = self.nll
        else:
            raise NotImplementedError("Loss function not implemented")

    def start_training(self, input_data: dict, config: dict, **kwargs: dict) -> dict:  # pylint: disable=R0915
        """
        This function starts the training of the QGAN.

        :param input_data: Dictionary with the variables from the circuit needed to start the training
        :param config: Training settings
        :param kwargs: Optional additional arguments
        :return: Dictionary including the solution
        """
        self.setup_training(input_data, config)
        generator_losses = []
        discriminator_losses = []
        best_kl_divergence = float('inf')
        best_generator_params = None
        pmfs_model = None
        best_sample = None

        n_batches = len(self.dataloader)

        for epoch in range(self.n_epochs):
            for batch, data in enumerate(self.dataloader):
                # Training the discriminator
                # Data from real distribution for training the discriminator
                real_data = data.float().to(self.device)
                self.discriminator.zero_grad()
                out_d_real = self.discriminator(real_data).view(-1)
                err_d_real = self.criterion(out_d_real, self.real_labels)
                err_d_real.backward()

                # Use Quantum Variational Circuit to generate fake samples
                fake_data, _ = self.generator.execute(self.params, self.batch_size)
                fake_data = fake_data.float().to(self.device)
                out_d_fake = self.discriminator(fake_data).view(-1)
                err_d_fake = self.criterion(out_d_fake, self.fake_labels)
                err_d_fake.backward()

                err_d = err_d_real + err_d_fake
                self.optimizer_discriminator.step()

                out_d_fake = self.discriminator(fake_data).view(-1)
                err_g = self.criterion(out_d_fake, self.real_labels)
                fake_data, _ = self.generator.execute(self.params, self.batch_size)
                gradients = self.generator.compute_gradient(
                    self.params,
                    self.discriminator,
                    self.criterion,
                    self.real_labels,
                    self.device
                )

                updated_params = self.params - self.learning_rate_generator * gradients
                self.params = updated_params

                self.discriminator_weights = self.discriminator.state_dict()
                generator_losses.append(err_g.item())
                discriminator_losses.append(err_d.item())

                # Calculate loss
                _, pmfs_model = self.generator.execute(self.params, self.n_shots)
                pmfs_model = np.asarray(pmfs_model.copy())

                loss = self.loss_func(pmfs_model[None,], self.target)
                self.accuracy.append(loss)

                self.writer.add_scalar("metrics/KL", loss, epoch * n_batches + batch)
                circuit_evals = (epoch * n_batches + batch) * self.batch_size * (2 * self.n_params + 1)
                self.writer.add_scalar("metrics/KL_circuit_evals", loss, circuit_evals)

                # Calculate and log the loss values at the end of each epoch
                self.writer.add_scalar('Loss/GAN_Generator', err_g.item(), circuit_evals)
                self.writer.add_scalar('Loss/GAN_Discriminator', err_d.item(), circuit_evals)

                if loss < best_kl_divergence:
                    best_kl_divergence = loss
                    best_generator_params = self.params.copy()  # Make a copy of the parameters
                _, best_pdf = self.generator.execute(best_generator_params, self.n_shots)
                best_pdf = np.asarray(best_pdf)
                best_pdf = best_pdf / best_pdf.sum()
                best_sample = self.sample_from_pmf(pmf=best_pdf, n_shots=self.n_shots)

                # Log the training progress
                log_message = (
                    f"Epoch: {epoch + 1}/{self.n_epochs}, "
                    f"Batch: {batch + 1}/{len(self.bins_train) // self.batch_size}, "
                    f"Discriminator Loss: {err_d.item()}, Generator Loss: {err_g.item()}, KL Divergence: {loss} "
                )

                logging.info(log_message)

            fig, ax = plt.subplots()
            ax.imshow(
                pmfs_model.reshape((2 ** (self.n_qubits // 2), 2 ** (self.n_qubits // 2))),
                cmap='binary',
                interpolation='none'
            )
            ax.set_title(f'Iteration {epoch}')
            self.writer.add_figure('grid_figure', fig, global_step=epoch)

            ax.clear()
            ax.imshow(
                self.target.reshape((2 ** (self.n_qubits // 2), 2 ** (self.n_qubits // 2))),
                cmap='binary',
                interpolation='none'
            )
            # Log the figure in TensorBoard
            ax.set_title("train")
            self.writer.add_figure("train", fig)

            # Plot the generator and discriminator losses on the existing figure
            ax.clear()
            ax.plot(generator_losses, label='Generator Loss', color='blue')
            ax.plot(discriminator_losses, label='Discriminator Loss', color='red')
            ax.legend()

            # Save the updated loss plot to TensorBoard
            self.writer.add_figure('Loss_Plot', fig, global_step=epoch)

        plt.close()
        self.writer.flush()
        self.writer.close()

        input_data["best_parameter"] = best_generator_params
        input_data["best_sample"] = best_sample
        input_data["KL"] = self.accuracy
        input_data["generator_loss"] = generator_losses
        input_data["discriminator_loss"] = discriminator_losses

        return input_data


class Discriminator(nn.Module):
    """
    This class defines the discriminator of the QGAN.
    """

    def __init__(self, input_length: int):
        super().__init__()
        self.dense1 = nn.Linear(int(input_length), 2 * int(input_length))
        self.dense2 = nn.Linear(2 * int(input_length), 1)

    def forward(self, x: torch.Tensor) -> float:
        """
        Initializes the weight tensor of the linear layers with values using a Xavier uniform distribution.

        :param x: Input of the discriminator
        :type x: torch.Tensor
        :return: Probability fake/real sample
        :rtype: float
        """
        h = funct.leaky_relu(self.dense1(x))
        h = funct.leaky_relu(self.dense2(h))
        return funct.sigmoid(h)

    @staticmethod
    def weights_init(m: nn.Linear) -> None:
        """
        Initializes the weight tensor of the linear
        layers with values using a Xavier uniform distribution.

        :param m: Neural network layer
        """
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight.data, gain=10)
            nn.init.constant_(m.bias.data, 1)


class QuantumGenerator:
    """
    This class defines the generator of the QGAN.
    """

    def __init__(self, n_qubits, execute_circuit, batch_size):
        self.n_qubits = n_qubits
        self.execute_circuit = execute_circuit
        self.batch_size = batch_size

    def execute(self, params: np.ndarray, n_shots: int) -> tuple[torch.Tensor, np.ndarray]:
        """
        Forward pass of the generator.

        :param params: Parameters of the quantum circuit
        :param n_shots: Number of shots
        :return: samples and the probability distribution generated by the quantum circuit
        """

        # Call the quantum circuit and obtain probability distributions
        pdfs, _ = self.execute_circuit(np.expand_dims(params, axis=0))
        pdfs = pdfs.flatten()

        # Sample from the provided probability distribution for the specified batch size
        sampling = torch.multinomial(torch.from_numpy(pdfs), n_shots, replacement=True)

        # Convert the sampling tensor to a list of integers
        binary_samples = [(sampling >> i) & 1 for i in range(self.n_qubits)]
        binary_samples = binary_samples[::-1]  # Reverse the order to match your expected format

        # Convert binary samples to a PyTorch tensor
        samples = torch.stack(binary_samples, dim=1).float()

        return samples, pdfs

    # pylint: disable=R0917
    def compute_gradient(self, params: np.ndarray, discriminator: torch.nn.Module, criterion: callable,
                         label: torch.Tensor, device: str) -> np.ndarray:
        """
        This function defines the forward pass of the generator.

        :param params: Parameters of the quantum circuit
        :param discriminator: Discriminator of the QGAN
        :param criterion: Loss function
        :param label: Label indicating of sample is true or fake
        :param device: Torch device (e.g., CPU or CUDA)
        :return: Samples and the probability distribution generated by the quantum circuit
        """
        shift = 0.5 * np.pi
        gradients = np.zeros(len(params))  # Initialize gradients as an array of zeros

        for i in range(len(params)):
            # Compute shifts for the i-th parameter
            positive_shifted_params = params.copy()
            negative_shifted_params = params.copy()
            positive_shifted_params[i] += shift
            negative_shifted_params[i] -= shift

            # Generate samples with shifts
            positive_samples, _ = self.execute(positive_shifted_params, self.batch_size)
            negative_samples, _ = self.execute(negative_shifted_params, self.batch_size)

            # Convert positive_samples and negative_samples to tensors
            positive_samples = positive_samples.to(device)
            negative_samples = negative_samples.to(device)

            # Compute discriminator outputs for all samples
            forward_outputs = discriminator(positive_samples).view(-1)
            backward_outputs = discriminator(negative_samples).view(-1)

            # Compute criterion differences for all samples
            forward_diff = criterion(forward_outputs, label)
            backward_diff = criterion(backward_outputs, label)

            # Calculate the gradient for the i-th parameter
            gradients[i] = 0.5 * (forward_diff.item() - backward_diff.item())

        return gradients



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pickle
import os
from abc import ABC
from qiskit import qpy

import numpy as np

from quark.modules.Core import Core
from quark.modules.applications.qml.DataHandler import DataHandler
from quark.utils import start_time_measurement, end_time_measurement


class DataHandlerGenerative(Core, DataHandler, ABC):
    """
    The task of the DataHandler module is to translate the application’s data
    and problem specification into preprocessed format.
    """

    def __init__(self, name: str):
        """
        Constructor method.
        """
        super().__init__()
        self.dataset_name = name
        self.generalization_mark = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "pandas", "version": "2.2.3"},
            {"name": "tensorboard", "version": "2.18.0"}
        ]

    def preprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[any, float]:
        """
        In this module, the preprocessing step is transforming the data to the correct target format.

        :param input_data: Collected information of the benchmarking process
        :param config: Config specifying the parameters of the training
        :param kwargs: Optional additional settings
        :return: Tuple with transformed problem and the time it took to map it
        """
        start = start_time_measurement()
        output = self.data_load(input_data, config)

        if "generalization_metrics" in list(output.keys()):
            self.generalization_mark = True

        return output, end_time_measurement(start)

    def postprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[dict, float]:
        """
        In this module, the postprocessing step is transforming the data to the correct target format.

        :param input_data: Original data
        :param config: Config specifying the parameters of the training
        :param kwargs: Optional additional settings
        :return: Tuple with an output_dictionary and the time it took
        """
        start = start_time_measurement()
        store_dir_iter = input_data["store_dir_iter"]
        current_directory = os.getcwd()

        if "samples_complete" in list(input_data.keys()):
            with open(f"{store_dir_iter}/samples_complete_{kwargs['rep_count']}.pkl", 'wb') as f:
                pickle.dump(input_data.pop("samples_complete"), f)

        evaluation, _ = self.evaluate(input_data)

        if self.generalization_mark is not None:
            self.metrics.add_metric_batch({"KL_best": evaluation["KL_best"]})
            metrics, _ = self.generalization()

            # Save generalization metrics
            with open(f"{store_dir_iter}/record_gen_metrics_{kwargs['rep_count']}.pkl", 'wb') as f:
                pickle.dump(metrics, f)

            self.metrics.add_metric_batch({"generalization_metrics": metrics})

        else:
            self.metrics.add_metric_batch({"KL_best": evaluation})

        # Save metrics per iteration
        if "inference" not in input_data.keys():
            self.tb_to_pd(logdir=store_dir_iter, rep=str(kwargs['rep_count']))
            self.metrics.add_metric_batch(
                {"metrics_pandas": os.path.relpath(f"{store_dir_iter}/data.pkl", current_directory)}
            )

            if self.generalization_mark is not None:
                np.save(f"{store_dir_iter}/histogram_generated.npy", evaluation["histogram_generated"])
            else:
                if "best_sample" in list(input_data.keys()):
                    samples = input_data["best_sample"]
                    n_shots = np.sum(samples)
                    histogram_generated = np.asarray(samples) / n_shots
                    histogram_generated[histogram_generated == 0] = 1e-8
                else:
                    histogram_generated = input_data["histogram_generated"]
                np.save(f"{store_dir_iter}/histogram_generated.npy", histogram_generated)
            self.metrics.add_metric_batch({"histogram_generated": os.path.relpath(
                f"{store_dir_iter}/histogram_generated.npy_{kwargs['rep_count']}.npy", current_directory)}
            )

            # Save histogram generated dataset
            np.save(f"{store_dir_iter}/histogram_train.npy", input_data.pop("histogram_train"))
            self.metrics.add_metric_batch({"histogram_train": os.path.relpath(
                f"{store_dir_iter}/histogram_train.npy_{kwargs['rep_count']}.npy", current_directory)}
            )

            # Save best parameters
            np.save(f"{store_dir_iter}/best_parameters_{kwargs['rep_count']}.npy", input_data.pop("best_parameter"))
            self.metrics.add_metric_batch({"best_parameter": os.path.relpath(
                f"{store_dir_iter}/best_parameters_{kwargs['rep_count']}.npy", current_directory)}
            )

            # Save training results
            input_data.pop("circuit_transpiled")
            with open(f"{store_dir_iter}/training_results-{kwargs['rep_count']}.pkl", 'wb') as f:
                pickle.dump(input_data, f)

            if "circuit_transpiled" in list(input_data.keys()):
                with open(f"{store_dir_iter}/transpiled_circuit_{kwargs['rep_count']}.qpy", 'wb') as f:
                    qpy.dump(input_data.pop("circuit_transpiled"), f)

            # Save variables transformed space
            if "Transformation" in list(input_data.keys()):
                self.metrics.add_metric_batch({"KL_best": input_data["KL_best_transformed"]})
                np.save(f"{store_dir_iter}/histogram_train_original.npy", input_data.pop("histogram_train_original"))
                np.save(f"{store_dir_iter}/histogram_generated_original.npy",
                        input_data.pop("histogram_generated_original"))
                with open(f"{store_dir_iter}/best_samples_transformed_{kwargs['rep_count']}.pkl", 'wb') as f:
                    pickle.dump(input_data["transformed_samples"], f)

        return input_data, end_time_measurement(start)

    def generalization(self) -> tuple[dict, float]:
        """
        Computes generalization metrics.

        :return: Evaluation and the time it took to create it
        """
        # Compute your metrics here
        metrics = {}  # Replace with actual metric calculations
        time_taken = 0.0  # Replace with actual time calculation
        return metrics, time_taken



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from quark.utils import _get_instance_with_sub_options


class Mapping(ABC):
    """
    The task of the mapping module is to translate the application’s data and problem specification
    into a mathematical formulation suitable for a solver.
    """

    def __init__(self):
        """
        Constructor method.
        """
        self.solver_options = []
        self.sub_options = None
        super().__init__()

    @abstractmethod
    def map(self, problem: any, config: dict) -> tuple[any, float]:
        """
        Maps the given problem into a specific format a solver can work with. E.g. graph to QUBO.

        :param problem: Problem instance which should be mapped to the target representation
        :param config: Instance of class Config specifying the mapping settings
        :return: The mapped problem and the time it took to create the mapping
        """
        pass

    def reverse_map(self, solution: any) -> tuple[any, float]:
        """
        Maps the solution back to the original problem. This might not be necessary in all cases, so the default is
        to return the original solution. This might be needed to convert the solution to a representation needed
        for validation and evaluation.

        :param solution: Solution to be reversed back to its original representation
        :return: Mapped solution and the time it took to create it
        """
        return solution, 0

    @abstractmethod
    def get_parameter_options(self) -> dict:
        """
        Method to return the parameters to fine tune the mapping.

        Should always be in this format:

        .. code-block:: json

            {
               "parameter_name":{
                  "values":[1, 2, 3],
                  "description":"How to scale your Lagrangian?"
               }
            }

        :return: Returns the available parameter options of this mapping
        """
        pass

    def get_submodule(self, solver_option: str) -> any:
        """
        If self.sub_options is not None, a solver is instantiated according to the information given in sub_options.
        Otherwise, get_solver is called as fall back.

        :param solver_option: The option for the solver
        :return: Instance of a solver class
        """
        if self.sub_options is None:
            return self.get_solver(solver_option)
        else:
            return _get_instance_with_sub_options(self.sub_options, solver_option)

    @abstractmethod
    def get_solver(self, solver_option: str) -> any:
        """
        Returns the default solver for a given string. This applies only if
        self.sub_options is None. See get_submodule.

        :param solver_option: desired solver option
        :return: Instance of solver class
        """
        pass

    def get_available_solver_options(self) -> list:
        """
        Returns all available solvers.

        :return: List of solvers
        """
        if self.sub_options is None:
            return self.solver_options
        else:
            return [option["name"] for option in self.sub_options]



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import random

import networkx as nx
import pulser

# define R_rydberg
R_rydberg = 9.75


def generate_hexagonal_graph(n_nodes: int, spacing: float, filling_fraction: float = 1.0) -> nx.Graph:
    """
    Generate a hexagonal graph layout based on the number of nodes and spacing.

    :param n_nodes: The number of nodes in the graph
    :param spacing: The spacing between nodes (atoms)
    :param filling_fraction: The fraction of available places in the lattice to be filled with nodes. (default: 1.0)
    :return: Networkx Graph representing the hexagonal graph layout
    """
    if not 0.0 < filling_fraction <= 1.0:
        raise ValueError("The filling fraction must be in the domain of (0.0, 1.0].")

    # Create a layout large enough to contain the desired number of atoms at the filling fraction
    n_traps = int(n_nodes / filling_fraction)
    hexagonal_layout = pulser.register.special_layouts.TriangularLatticeLayout(
        n_traps=n_traps, spacing=spacing
    )

    # Fill the layout with traps
    reg = hexagonal_layout.hexagonal_register(n_traps)
    ids = reg._ids  # pylint: disable=W0212
    coords = [coord.tolist() for coord in reg._coords]  # pylint: disable=W0212
    traps = dict(zip(ids, coords))

    # Remove random atoms to get the desired number of atoms
    while len(traps) > n_nodes:
        atom_to_remove = random.choice(list(traps))
        traps.pop(atom_to_remove)

    # Rename the atoms
    node_positions = {i: traps[trap] for i, trap in enumerate(traps.keys())}  # pylint: disable=C0206

    # Create the graph
    hexagonal_graph = nx.Graph()

    # Add nodes to the graph
    for node_id, coord in node_positions.items():
        hexagonal_graph.add_node(node_id, pos=coord)

    # Generate the edges and add them to the graph
    edges = _generate_edges(node_positions=node_positions)
    hexagonal_graph.add_edges_from(edges)

    return hexagonal_graph


def _generate_edges(node_positions: dict[list[int, list[float]]], radius: float = R_rydberg) -> list[tuple]:
    """
    Generate edges between vertices within a given distance 'radius', which defaults to R_rydberg.

    :param node_positions: A dictionary with the node ids as keys, and the node coordinates as values
    :param radius: When the distance between two nodes is smaller than this radius, an edge is generated between them
    :return: A list of 2-tuples. Each 2-tuple contains two different node ids and represents an edge between those nodes
    """
    edges = []
    vertex_keys = list(node_positions.keys())
    for i, vertex_key in enumerate(vertex_keys):
        for neighbor_key in vertex_keys[i + 1:]:
            distance = _vertex_distance(node_positions[vertex_key], node_positions[neighbor_key])
            if distance <= radius:
                edges.append((vertex_key, neighbor_key))
    return edges


def _vertex_distance(v0: tuple[float, ...], v1: tuple[float, ...]) -> float:
    """
    Calculates distance between two n-dimensional vertices.
    For 2 dimensions: distance = sqrt((x0 - x1)**2 + (y0 - y1)**2)

    :param v0: Coordinates of the first vertex
    :param v1: Coordinates of the second vertex
    return: Distance between the vertices
    """
    squared_difference = sum((coordinate0 - coordinate1) ** 2 for coordinate0, coordinate1 in zip(v0, v1))

    return math.sqrt(squared_difference)
